import math
from calendar import month_abbr
from datetime import datetime
from functools import cached_property

import numpy as np
import pandas as pd
from pandas.testing import assert_index_equal

from . import fields as fld
from . import utils
from .reader import _SETDataReaderCached

nan = float("nan")

summary_columns = [
    "timestamp",
    "port_value_with_dividend",
    "port_value",
    "total_market_value",
    "cash",
    "cashflow",
    "dividend",
    "cumulative_dividend",
    "commission",
]
position_columns = [
    "timestamp",
    "symbol",
    "volume",
    "cost_price",
    "close_price",
    "close_value",
    "pct_profit",
]
trade_columns = [
    "matched_at",
    "symbol",
    "side",
    "volume",
    "price",
    "commission",
]
dividend_columns = [
    "ex_date",
    "symbol",
    "dps",
    "volume",
    "amount",
    "before_ex_date",
    "pay_date",
]
summary_trade_columns = [
    "exit_at",
    "symbol",
    "cost_price",
    "sell_price",
    "volume",
    "commission",
    "return",
    "pct_return",
    "entry_at",
    "hold_days",
]


def return_nan_on_failure(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ZeroDivisionError:
            return nan

    return wrapper


class SETBacktestReport:
    def __init__(
        self,
        initial_capital: float,
        pct_commission: float,
        pct_buy_slip: float,
        pct_sell_slip: float,
        cash_series: pd.Series,
        position_df: pd.DataFrame,
        trade_df: pd.DataFrame,
    ):
        """SETBacktestReport.

        Parameters
        ----------
        initial_capital : float
            initial capital.
        pct_commission : float
            percent commission.
        pct_buy_slip : float
            percent of buy price increase.
        pct_sell_slip : float
            percent of sell price decrease.
        cash_series : pd.Series
            series of cash.
        position_df : pd.DataFrame
            dataframe of position.
                - timestamp
                - symbol
                - volume
                - cost_price
        trade_df : pd.DataFrame
            dataframe of trade.
                - matched_at
                - symbol
                - volume
                - price
                - pct_commission
        """
        self._initial_capital = initial_capital
        self._pct_commission = pct_commission
        self._pct_buy_slip = pct_buy_slip
        self._pct_sell_slip = pct_sell_slip

        self._cash_series = cash_series
        self._position_df = position_df
        self._trade_df = trade_df

        self._sdr = _SETDataReaderCached()

    @cached_property
    def summary_df(self) -> pd.DataFrame:
        """Summary DataFrame.

        Returns
        -------
        pd.DataFrame
            - timestamp
            - port_value_with_dividend
            - port_value
            - total_market_value
            - cash
            - cashflow
            - dividend
            - cumulative_dividend
            - commission
        """
        df = self._cash_series.to_frame("cash")

        df["cashflow"] = df["cash"] - df["cash"].shift(
            1, fill_value=self.initial_capital
        )

        df["commission"] = (
            self.trade_df.set_index("matched_at")["commission"].groupby(level=0).sum()
        )
        df["commission"] = df["commission"].fillna(0.0)

        df["total_market_value"] = (
            self.position_df.set_index("timestamp")["close_value"]
            .groupby(level=0)
            .sum()
        )
        df["total_market_value"] = df["total_market_value"].fillna(0.0)

        df["port_value"] = df["total_market_value"] + df["cash"]

        # pay_date can be non trade date.
        dividend_df = self.dividend_df.copy()
        dividend_df["pay_date"] += self._sdr._SETBusinessDay(0)  # type: ignore
        df["dividend"] = (
            dividend_df.set_index("pay_date")["amount"].groupby(level=0).sum()
        )
        df["dividend"] = df["dividend"].fillna(0.0)

        df["cumulative_dividend"] = df["dividend"].cumsum()

        df["port_value_with_dividend"] = df["port_value"] + df["cumulative_dividend"]

        df.index.name = "timestamp"
        df = df.reset_index()

        # sort column
        df = df[summary_columns]

        return df

    @cached_property
    def position_df(self) -> pd.DataFrame:
        """SETPosition DataFrame.

        Returns
        -------
        pd.DataFrame
            - timestamp
            - symbol
            - volume
            - cost_price
            - close_price
            - close_value
            - pct_profit
        """
        df = self._position_df.copy()

        if df.empty:
            df = pd.DataFrame(columns=position_columns)
            df["timestamp"] = pd.to_datetime(df["timestamp"])
            return df

        df["close_value"] = df["close_price"] * df["volume"]
        df["pct_profit"] = (df["close_price"] / df["cost_price"]) - 1.0

        # sort column
        df = df[position_columns]

        return df

    @cached_property
    def trade_df(self) -> pd.DataFrame:
        """Trade DataFrame.

        Returns
        -------
        pd.DataFrame
            - matched_at
            - symbol
            - side
            - volume
            - price
            - commission
        """
        df = self._trade_df.copy()
        df["side"] = df["volume"].apply(
            lambda x: fld.SIDE_BUY if x > 0 else fld.SIDE_SELL
        )
        df["volume"] = df["volume"].abs()
        df["commission"] = df["price"] * df["volume"] * df["pct_commission"]
        df = df.drop(columns=["pct_commission"])

        # sort column
        df = df[trade_columns]

        return df

    @cached_property
    def dividend_df(self) -> pd.DataFrame:
        """Dividend DataFrame.

        Returns
        -------
        pd.DataFrame
            - ex_date
            - symbol
            - dps
            - volume
            - amount
            - before_ex_date
            - pay_date
        """
        position_df = self.position_df.copy()

        if position_df.empty:
            df = pd.DataFrame(columns=dividend_columns)
            df["ex_date"] = pd.to_datetime(df["ex_date"])
            df["before_ex_date"] = pd.to_datetime(df["before_ex_date"])
            df["pay_date"] = pd.to_datetime(df["pay_date"])
            return df

        # Get cash dividend dataframe
        symbol_list = position_df["symbol"].unique().tolist()
        start_date = utils.date_to_str(position_df["timestamp"].min())
        end_date = utils.date_to_str(position_df["timestamp"].max())
        cash_dvd_df = self._sdr.get_dividend(
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
            ca_type_list=["CD"],
        )

        cash_dvd_df["pay_date"] = cash_dvd_df["pay_date"].fillna(cash_dvd_df["ex_date"])

        cash_dvd_df["before_ex_date"] = cash_dvd_df["ex_date"] - self._sdr._SETBusinessDay(1)  # type: ignore

        position_df = position_df.rename(columns={"timestamp": "before_ex_date"})

        df = cash_dvd_df.merge(
            position_df, how="inner", on=["before_ex_date", "symbol"], sort=True
        )

        df["amount"] = df["dps"] * df["volume"]

        # sort column
        df = df[dividend_columns]

        return df

    @cached_property
    def stat_df(self) -> pd.DataFrame:
        """Stat DataFrame.

        Returns
        -------
        pd.DataFrame
            columns - nav names
            indexes
                - pct_net_profit
                - cagr
                - pct_maximum_drawdown
                - cagr_divided_maxdd
                - pct_win_per_trade
                - std
                - cagr_divided_std
                - pct_exposure
                - total_commission
                - initial_capital
                - ending_capital
                - net_profit
                - maximum_drawdown
                - all_trades
                - avg_profit_loss
                - pct_avg_profit_loss
                - avg_bar_held
                - win_trades
                - total_profit
                - avg_profit
                - pct_avg_profit
                - avg_win_bar_held
                - max_win_consecutive
                - loss_trades
                - total_loss
                - avg_loss
                - pct_avg_loss
                - avg_lose_bar_held
                - max_lose_consecutive
                - start_date
                - end_date
                - pct_commission
                - pct_buy_slip
                - pct_sell_slip
        """
        stat_dict = {
            # Highlight
            "pct_net_profit": self.pct_net_profit,
            "cagr": self.cagr,
            "pct_maximum_drawdown": self.pct_maximum_drawdown,
            "cagr_divided_maxdd": self.cagr_divided_maxdd,
            "pct_win_per_trade": self.pct_win_per_trade,
            "std": self.std,
            "cagr_divided_std": self.cagr_divided_std,
            "pct_exposure": self.pct_exposure,
            "total_commission": self.total_commission,
            # Stat
            "initial_capital": self.initial_capital,
            "ending_capital": self.ending_capital,
            "net_profit": self.net_profit,
            "maximum_drawdown": self.maximum_drawdown,
            "all_trades": self.all_trades,
            "avg_profit_loss": self.avg_profit_loss,
            "pct_avg_profit_loss": self.pct_avg_profit_loss,
            "avg_bar_held": self.avg_bar_held,
            "win_trades": self.win_trades,
            "total_profit": self.total_profit,
            "avg_profit": self.avg_profit,
            "pct_avg_profit": self.pct_avg_profit,
            "avg_win_bar_held": self.avg_win_bar_held,
            "max_win_consecutive": self.max_win_consecutive,
            "loss_trades": self.loss_trades,
            "total_loss": self.total_loss,
            "avg_loss": self.avg_loss,
            "pct_avg_loss": self.pct_avg_loss,
            "avg_lose_bar_held": self.avg_lose_bar_held,
            "max_lose_consecutive": self.max_lose_consecutive,
            # Setup
            "start_date": self.start_date,
            "end_date": self.end_date,
            "pct_commission": self.pct_commission,
            "pct_buy_slip": self.pct_buy_slip,
            "pct_sell_slip": self.pct_sell_slip,
        }

        df = pd.DataFrame(stat_dict).T

        # sort columns
        df = df[self._nav_df.columns]

        return df

    @cached_property
    def summary_trade_df(self) -> pd.DataFrame:
        """Summary Trade DataFrame. Each row is sell.

        Returns
        -------
        pd.DataFrame
            - exit_at
            - symbol
            - cost_price
            - sell_price
            - volume
            - commission
            - return
            - pct_return
            - entry_at
            - hold_days
        """
        trade_df = self.trade_df.copy()

        # avg cost price
        trade_df = self._summary_trade_cost_price(trade_df)

        # sell all in position
        df = self._summary_trade_sell_all_position()
        assert_index_equal(df.columns.sort_values(), trade_df.columns.sort_values())
        trade_df = pd.concat([trade_df, df], ignore_index=True)

        # datetime in
        trade_df = self._summary_trade_entry_at(trade_df)

        df = trade_df[trade_df["side"] == fld.SIDE_SELL]
        df = df.reset_index(drop=True)

        # sell price
        df = df.rename(columns={"price": "sell_price", "matched_at": "exit_at"})

        # commission from buy and sell
        df["commission"] = (
            (df["sell_price"] + df["cost_price"]) * df["volume"] * self._pct_commission
        ).abs()

        # return
        df["return"] = ((df["sell_price"] - df["cost_price"]) * df["volume"]) - df[
            "commission"
        ]

        # pct_return
        df["pct_return"] = df["return"] / df["cost_price"] / df["volume"]

        # hold days
        df["hold_days"] = (df["exit_at"] - df["entry_at"]).dt.days

        return df[summary_trade_columns]

    @cached_property
    def cumulative_return_df(self) -> pd.DataFrame:
        """Cumulative Return DataFrame.

        Returns
        -------
        pd.DataFrame
            index is trade date, columns is nav names
        """
        return self._nav_df / self.initial_capital

    @cached_property
    def monthly_return_df(self) -> pd.DataFrame:
        """Monthly Return DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        df = self._nav_df
        init_cap = self.initial_capital

        monthly_return = self._return_by_period(df, init_cap=init_cap, period="M")
        yearly_return = self._return_by_period(df, init_cap=init_cap, period="Y")

        out = monthly_return.merge(
            yearly_return,
            how="inner",
            left_index=True,
            right_index=True,
            validate="1:1",
        )

        return out

    @cached_property
    def price_distribution_df(self) -> pd.DataFrame:
        """Price Distribution DataFrame.

        Returns
        -------
        pd.DataFrame

        Examples
        --------
        >>>                     frequency
        ... pct_return
        ... (-0.25, -0.2]             153
        ... (-0.2, -0.15]             352
        ... (-0.15, -0.1]             459
        ... (-0.1, -0.05]             698
        ... (-0.05, -0.0]            1268
        ... (-0.0, 0.05]             1256
        ... (0.05, 0.1]               571
        ... (0.1, 0.15]               374
        ... (0.15, 0.2]               181
        """
        pct_return = self.summary_trade_df["pct_return"]

        # bins
        step = 0.05

        min_r = pct_return.min()
        max_r = pct_return.max()

        if np.isnan(min_r) and np.isnan(max_r):
            min_r, max_r = 0, 0
        else:
            if min_r % step == 0:
                min_r -= step
            else:
                min_r = math.ceil(min_r / step - 1) * step
            max_r += step

        bins = np.arange(min_r, max_r, step)

        # histogram
        price_dis = pct_return.groupby(pd.cut(pct_return, bins)).count()

        return price_dis.to_frame("frequency")

    @cached_property
    def drawdown_percent_df(self) -> pd.DataFrame:
        """Drawdown Percent DataFrame.

        Returns
        -------
        pd.DataFrame
            index is trade date, columns is nav names
        """
        return (self._nav_df / self._nav_df.cummax()) - 1

    def to_excel(self, path: str):
        """Export to Excel.

        Parameters
        ----------
        path : str
            Path to Excel file.
        """
        with pd.ExcelWriter(
            path, engine="xlsxwriter", datetime_format="YYYY-MM-DD"
        ) as writer:
            self.summary_df.to_excel(writer, sheet_name="summary", index=False)
            self.position_df.to_excel(writer, sheet_name="position", index=False)
            self.trade_df.to_excel(writer, sheet_name="trade", index=False)
            self.stat_df.to_excel(writer, sheet_name="stat")
            self.summary_trade_df.to_excel(
                writer, sheet_name="summary_trade", index=False
            )
            self.cumulative_return_df.to_excel(writer, sheet_name="cumulative_return")
            self.monthly_return_df.to_excel(writer, sheet_name="monthly_return")
            self.dividend_df.to_excel(writer, sheet_name="dividend", index=False)
            self.price_distribution_df.to_excel(writer, sheet_name="price_distribution")
            self.drawdown_percent_df.to_excel(writer, sheet_name="drawdown_percent")

    """
    Stat
    """

    @property
    @return_nan_on_failure
    def pct_net_profit(self) -> pd.Series:
        """Net profit divided by initial capital."""
        return self.net_profit / self.initial_capital

    @property
    @return_nan_on_failure
    def cagr(self) -> pd.Series:
        """Compound annual growth rate (CAGR)"""
        return (self.ending_capital / self.initial_capital) ** (1 / self._n_year) - 1

    @property
    @return_nan_on_failure
    def pct_maximum_drawdown(self) -> pd.Series:
        """Percent maximum drawdown."""
        nav = self._nav_df
        return (nav / nav.cummax() - 1).min()

    @property
    @return_nan_on_failure
    def cagr_divided_maxdd(self) -> pd.Series:
        """Compound Annual % Return divided by Max.

        system % drawdown.
        """
        return self.cagr / self.pct_maximum_drawdown.abs()

    @property
    @return_nan_on_failure
    def pct_win_per_trade(self) -> float:
        """Percent of win trades."""
        return self.win_trades / self.all_trades

    @property
    @return_nan_on_failure
    def std(self) -> pd.Series:
        """Standard deviation of profit/loss."""
        nav = self._nav_df
        return_per_day = nav / nav.shift(1) - 1
        trade_date_per_year = nav.shape[0] / self._n_year
        return return_per_day.std() * (trade_date_per_year**0.5)

    @property
    @return_nan_on_failure
    def cagr_divided_std(self) -> pd.Series:
        """Compound Annual % Return divided by Standard deviation."""
        return self.cagr / self.std

    @property
    @return_nan_on_failure
    def pct_exposure(self) -> pd.Series:
        """Percent of exposure."""
        df = self.summary_df.copy()
        df["portfolio"] = df["total_market_value"] / df["port_value"]
        df["portfolio_with_dividend"] = (
            df["total_market_value"] / df["port_value_with_dividend"]
        )
        return df[["portfolio", "portfolio_with_dividend"]].mean()

    @property
    @return_nan_on_failure
    def total_commission(self) -> float:
        """Total commission."""
        return self.summary_df["commission"].sum()

    @property
    @return_nan_on_failure
    def initial_capital(self) -> float:
        """Initial capital."""
        return self._initial_capital

    @property
    @return_nan_on_failure
    def ending_capital(self) -> pd.Series:
        """Ending capital."""
        return self._nav_df.iloc[-1]

    @property
    @return_nan_on_failure
    def net_profit(self) -> pd.Series:
        """Net profit."""
        return self.ending_capital - self.initial_capital

    @property
    @return_nan_on_failure
    def maximum_drawdown(self) -> pd.Series:
        """Maximum drawdown."""
        nav = self._nav_df
        return (nav - nav.cummax()).min()

    @property
    @return_nan_on_failure
    def all_trades(self) -> int:
        """Total number of trades."""
        df = self.summary_trade_df
        return df.shape[0]

    @property
    @return_nan_on_failure
    def avg_profit_loss(self) -> pd.Series:
        """Average profit/loss.

        also known as Expectancy ($) - (Profit of winners + Loss of losers)/(number of trades), represents expected dollar gain/loss per trade
        """
        return self.net_profit / self.all_trades

    @property
    @return_nan_on_failure
    def pct_avg_profit_loss(self) -> float:
        """Percent average profit/loss.

        also known as Expectancy (%) - '(% Profit of winners + % Loss of losers)/(number of trades), represents expected percent gain/loss per trade
        """
        return (
            (self.pct_avg_profit * self.win_trades)
            + (self.pct_avg_loss * self.loss_trades)
        ) / self.all_trades

    @property
    @return_nan_on_failure
    def avg_bar_held(self) -> float:
        """sum of bars in trades / number of trades."""
        df = self.summary_trade_df
        return df["hold_days"].mean()

    @property
    @return_nan_on_failure
    def win_trades(self) -> int:
        """Number of win trades."""
        return self._is_win_trade.sum()

    @property
    @return_nan_on_failure
    def total_profit(self) -> float:
        """Total profit."""
        df = self.summary_trade_df
        df = df[self._is_win_trade]
        return df["return"].sum()

    @property
    @return_nan_on_failure
    def avg_profit(self) -> float:
        """Average profit."""
        df = self.summary_trade_df
        df = df[self._is_win_trade]
        return df["return"].mean()

    @property
    @return_nan_on_failure
    def pct_avg_profit(self) -> float:
        """Percent average profit."""
        df = self.summary_trade_df
        df = df[self._is_win_trade]
        return df["pct_return"].mean()

    @property
    @return_nan_on_failure
    def avg_win_bar_held(self) -> float:
        """Average win bar held."""
        df = self.summary_trade_df
        df = df[self._is_win_trade]
        return df["hold_days"].mean()

    @property
    @return_nan_on_failure
    def max_win_consecutive(self) -> int:
        """Maximum win consecutive."""
        s = self._is_win_trade
        return utils.count_true_consecutive(s)

    @property
    @return_nan_on_failure
    def loss_trades(self) -> int:
        """Number of loss trades."""
        return (~self._is_win_trade).sum()

    @property
    @return_nan_on_failure
    def total_loss(self) -> float:
        """Total loss."""
        df = self.summary_trade_df
        df = df[~self._is_win_trade]
        return df["return"].sum()

    @property
    @return_nan_on_failure
    def avg_loss(self) -> float:
        """Average loss."""
        df = self.summary_trade_df
        df = df[~self._is_win_trade]
        return df["return"].mean()

    @property
    @return_nan_on_failure
    def pct_avg_loss(self) -> float:
        """Percent average loss."""
        df = self.summary_trade_df
        df = df[~self._is_win_trade]
        return df["pct_return"].mean()

    @property
    @return_nan_on_failure
    def avg_lose_bar_held(self) -> float:
        """Average lose bar held."""
        df = self.summary_trade_df
        df = df[~self._is_win_trade]
        return df["hold_days"].mean()

    @property
    @return_nan_on_failure
    def max_lose_consecutive(self) -> int:
        """Maximum lose consecutive."""
        s = ~self._is_win_trade
        return utils.count_true_consecutive(s)

    @cached_property
    def start_date(self) -> datetime:
        """Start date."""
        out = self._nav_df.index[0]
        assert isinstance(out, datetime)
        return out

    @cached_property
    def end_date(self) -> datetime:
        """End date."""
        out = self._nav_df.index[-1]
        assert isinstance(out, datetime)
        return out

    @property
    def pct_commission(self) -> float:
        """Percent commission."""
        return self._pct_commission

    @property
    def pct_buy_slip(self) -> float:
        """Percent buy slip."""
        return self._pct_buy_slip

    @property
    def pct_sell_slip(self) -> float:
        """Percent sell slip."""
        return self._pct_sell_slip

    @property
    def _n_year(self) -> float:
        """Number of years."""
        return (self.end_date - self.start_date).days / 365

    @property
    def _is_win_trade(self) -> pd.Series:
        """Is win trade."""
        return self.summary_trade_df["return"] > 0

    """
    Protected
    """

    @cached_property
    def _nav_df(self) -> pd.DataFrame:
        return self.summary_df.set_index("timestamp")[
            ["port_value", "port_value_with_dividend"]
        ].rename(
            columns={
                "port_value": "portfolio",
                "port_value_with_dividend": "portfolio_with_dividend",
            }
        )

    def _summary_trade_cost_price(self, trade_df: pd.DataFrame) -> pd.DataFrame:
        """Add cost_price to trade_df."""
        pos_df = self.position_df[["timestamp", "symbol", "cost_price"]]
        pos_df = pos_df.rename(columns={"timestamp": "matched_at"})

        # get cost price from tomorrow position
        pos_df["matched_at"] += self._sdr._SETBusinessDay(1)  # type: ignore

        # set index for merge
        df = trade_df.merge(
            pos_df, how="left", on=["matched_at", "symbol"], validate="1:1"
        )

        return df

    def _summary_trade_sell_all_position(self) -> pd.DataFrame:
        """Add trade if sell all position.

        No commission. Sell at close price.
        """
        df = self.position_df.copy()

        df = df[df["timestamp"] == self.end_date]
        df = df.rename(columns={"close_price": "price"})
        df["side"] = fld.SIDE_SELL
        df["commission"] = 0
        df = df.drop(columns=["close_value", "pct_profit"])

        df = df.rename(columns={"timestamp": "matched_at"})

        df = df.reset_index(drop=True)

        return df

    @staticmethod
    def _summary_trade_entry_at(df: pd.DataFrame) -> pd.DataFrame:
        """Add entry_at to trade_df.

        datetime in is nearest buy.
        """
        df["entry_at"] = df.loc[df["side"] == fld.SIDE_BUY, "matched_at"]

        tmp = df.groupby(["symbol"]).fillna(method="pad")  # type: ignore
        df["entry_at"] = tmp["entry_at"]

        return df

    @staticmethod
    def _return_by_period(
        df: pd.DataFrame, init_cap: float, period: str
    ) -> pd.DataFrame:
        """Return by period.

        Parameters
        ----------
        df : pd.DataFrame
            Nav dataframe.
        init_cap : float
            Initial capital.
        period : str
            "M" (monthly) or "Y" (yearly)

        Returns
        -------
        pd.DataFrame
            Return by period.
        """
        df = df.resample(period).last()  # type: ignore
        df = df.sort_index()
        df = (df / df.shift(fill_value=init_cap)) - 1

        df = df.melt(var_name="name", value_name="value", ignore_index=False)

        df["year"] = df.index.year  # type: ignore
        df["month"] = df.index.month  # type: ignore

        if period == "M":
            df = pd.pivot_table(
                df, index=["name", "year"], columns="month", values="value"
            )
            df = df.rename(
                columns={i: month_abbr[i] for i in range(1, len(month_abbr))}
            )
            df = df.reindex(columns=month_abbr[1:])  # type: ignore
        else:
            df = pd.pivot_table(df, index=["name", "year"], values="value")
            df = df.rename(columns={"value": "YTD"})
            df.columns.name = "month"

        return df



from functools import lru_cache
from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from . import fields as fld
from . import utils
from .errors import InputError
from .indicators import TA
from .reader import _SETDataReaderCached


class SETSignalCreator:
    ta = TA

    def __init__(
        self,
        start_date: str = "2010-01-01",
        end_date: Optional[str] = None,
        index_list: List[str] = ["SET100"],
        symbol_list: List[str] = [],
    ):
        """Initialize SETSignalCreator.

        Symbol in universe is union of symbol_list and symbols in index_list.

        Parameters
        ----------
        start_date : str
            Start date of data.
        end_date : str
            End date of data.
        index_list : List[str]
            List of index name. by default ['SET100']
                - SET
                - mai
                - SETWB
                - SETTHSI
                - SETCLMV
                - SETHD
                - sSET
                - SET100
                - SET50
        symbol_list : List[str]
            List of symbol name. by default []
        """
        self._index_list: List[str] = [i.upper() for i in index_list]
        self._symbol_list: List[str] = [i.upper() for i in symbol_list]
        self._start_date: str = start_date
        self._end_date: Optional[str] = end_date

        self._sdr = _SETDataReaderCached()

    def get_data(
        self,
        field: str,
        timeframe: str,
        value_by: str = "stock",
        method: str = "constant",
        period: int = 1,
        shift: int = 0,
    ) -> pd.DataFrame:
        """Get DataFrame which columns are symbol in universe and index is
        trade date from start_date to end_date.

        Banned symbol will be replaced with nan.

        Parameters
        ----------
        field : str
            Name of data field depends on timeframe and value_by.
        timeframe : str
            - daily
            - quarterly
            - yearly
            - ttm
            - ytd
        value_by : str, default 'stock'
            - stock
            - industry
            - sector
        method : str, default 'constant'
            Name of Dataframe rolling window functions. See <https://pandas.pydata.org/docs/reference/window.html#rolling-window-functions> for details.
            - constant
            - count
            - sum
            - mean
            - median
            - var
            - std
            - min
            - max
            - corr
            - cov
            - skew
            - kurt
            - apply
            - aggregate
            - quantile
            - sem
            - rank
        period : int, default 1
            Number of periods for rolling Dataframe. Period must be greater than 0.
        shift : int, default 0
            Number of periods to shift Dataframe. Shift must be greater than or equal to 0.

        Returns
        -------
        pd.DataFrame
            - symbol : str as column
            - trade date : date as index

        Examples
        --------
        >>> from ezyquant import SETSignalCreator
        >>> ssc = SETSignalCreator(
        ...     start_date="2022-01-01",
        ...     end_date="2022-01-10",
        ...     index_list=[],
        ...     symbol_list=["COM7", "MALEE"],
        ... )
        >>> ssc.get_data(
        ...     field="close",
        ...     timeframe="daily",
        ...     value_by="stock",
        ...     method="constant",
        ...     period=0,
        ...     shift=0,
        ... )
                      COM7  MALEE
        2022-01-04  41.875   6.55
        2022-01-05  41.625   6.50
        2022-01-06  41.500   6.50
        2022-01-07  41.000   6.40
        2022-01-10  40.875   6.30
        ...            ...    ...
        2022-03-28  42.750   5.75
        2022-03-29  41.500   5.75
        2022-03-30  42.250   5.70
        2022-03-31  43.000   5.70
        2022-04-01  42.500   5.70

        >>> ssc.get_data(
        ...     field="cash",
        ...     timeframe="quarterly",
        ...     value_by="stock",
        ...     method="constant",
        ...     period=0,
        ...     shift=0,
        ... )
                          COM7     MALEE
        2022-01-04         NaN       NaN
        2022-01-05         NaN       NaN
        2022-01-06         NaN       NaN
        2022-01-07         NaN       NaN
        2022-01-10         NaN       NaN
        ...                ...       ...
        2022-03-28  1656882.22  80321.08
        2022-03-29  1656882.22  80321.08
        2022-03-30  1656882.22  80321.08
        2022-03-31  1656882.22  80321.08
        2022-04-01  1656882.22  80321.08
        """
        timeframe = timeframe.lower()
        value_by = value_by.lower()
        method = method.lower()

        # TODO: load mare data for shift

        if period < 1:
            raise InputError(f"period must be greater than 0. but {period} is given.")
        if shift < 0:
            raise InputError(
                f"shift must be greater than or equal to 0, but {shift} is given."
            )
        if method == fld.METHOD_CONSTANT and period != 1:
            raise InputError(f"{fld.METHOD_CONSTANT} method only support period=1")

        symbol_list = self._get_symbol_in_universe()

        if timeframe == fld.TIMEFRAME_DAILY:
            if value_by == fld.VALUE_BY_STOCK:
                df = self._sdr.get_data_symbol_daily(
                    field=field,
                    symbol_list=symbol_list,
                    start_date=self._start_date,
                    end_date=self._end_date,
                )

                # Forward fill 0 and NaN with prior
                if field in {
                    fld.D_OPEN,
                    fld.D_HIGH,
                    fld.D_LOW,
                    fld.D_CLOSE,
                    fld.D_AVERAGE,
                    fld.D_LAST_BID,
                    fld.D_LAST_OFFER,
                }:
                    prior_df = self._sdr.get_data_symbol_daily(
                        field=fld.D_PRIOR,
                        symbol_list=symbol_list,
                        start_date=self._start_date,
                        end_date=self._end_date,
                    )
                    prior_df = self._reindex_trade_date(prior_df)
                    prior_df = prior_df.replace(0, np.nan)
                    prior_df = prior_df.fillna(method="ffill")

                    df = self._reindex_trade_date(df)
                    df = df.replace(0, np.nan)
                    df = df.fillna(prior_df)

            elif value_by == fld.VALUE_BY_SECTOR:
                df = self._sdr._get_daily_sector_info_by_security(
                    field=field,
                    symbol_list=symbol_list,
                    start_date=self._start_date,
                    end_date=self._end_date,
                    f_data="S",
                )
            elif value_by == fld.VALUE_BY_INDUSTRY:
                df = self._sdr._get_daily_sector_info_by_security(
                    field=field,
                    symbol_list=symbol_list,
                    start_date=self._start_date,
                    end_date=self._end_date,
                    f_data="I",
                )
            else:
                raise InputError(
                    f"{value_by} is invalid value_by. Please read document to check valid value_by."
                )

            df = self._reindex_trade_date(df)

            df = df.shift(shift)

            if method != fld.METHOD_CONSTANT:
                df = self._rolling(df, method=method, period=period)

        elif timeframe in (
            fld.TIMEFRAME_QUARTERLY,
            fld.TIMEFRAME_YEARLY,
            fld.TIMEFRAME_TTM,
            fld.TIMEFRAME_YTD,
        ):
            if value_by == fld.VALUE_BY_STOCK:
                df = self._sdr._get_fundamental_data(
                    field=field,
                    symbol_list=symbol_list,
                    start_date=self._start_date,
                    end_date=self._end_date,
                    timeframe=timeframe,
                    fillna_value=np.inf,
                )

                df = df.apply(
                    lambda x: self._rolling_skip_na_keep_inf(
                        x, method=method, period=period, shift=shift
                    ),
                    axis=0,
                )
                df = self._reindex_trade_date(df)
                df = df.fillna(method="ffill")
                df = df.replace(np.inf, np.nan)
            else:
                raise InputError(
                    f"{value_by} is invalid value_by. Please read document to check valid value_by."
                )

        else:
            raise InputError(
                f"{timeframe} is invalid timeframe. Please read document to check valid timeframe."
            )

        df.index.freq = None  # type: ignore

        df = df.mask(self.is_banned(), np.nan)

        return df

    def is_universe(self, universe: str) -> pd.DataFrame:
        """Return Dataframe of boolean is universe.

        Parameters
        ----------
        universe : str
            Can be Market, Sector, Industry, or Index.

        Returns
        -------
        pd.DataFrame of boolean
            - symbol : str as column
            - trade date : date as index

        Examples
        --------
        >>> from ezyquant import SETSignalCreator
        >>> ssc = SETSignalCreator(
        ...     start_date="2022-01-01",
        ...     end_date="2022-01-10",
        ...     index_list=[],
        ...     symbol_list=["COM7", "MALEE"],
        ... )
        >>> ssc.is_universe("SET100")
                    COM7  MALEE
        2022-01-04  True  False
        2022-01-05  True  False
        2022-01-06  True  False
        2022-01-07  True  False
        2022-01-10  True  False
        """
        universe = universe.upper()

        try:
            return self._is_universe_static(universe)
        except InputError:
            return self._is_universe_dynamic(universe)

    def is_banned(self) -> pd.DataFrame:
        """Return Dataframe of boolean is banned by Delisted or Suspension
        (SP).

        Returns
        -------
        pd.DataFrame of boolean
            - symbol : str as column
            - trade date : date as index

        Examples
        --------
        >>> from ezyquant import SETSignalCreator
        >>> ssc = SETSignalCreator(
        ...     start_date="2022-01-01",
        ...     end_date="2022-01-10",
        ...     index_list=[],
        ...     symbol_list=["COM7", "MALEE", "THAI"],
        ... )
        >>> ssc.is_banned()
                     COM7  MALEE  THAI
        2022-01-04  False  False  True
        2022-01-05  False  False  True
        2022-01-06  False  False  True
        2022-01-07  False  False  True
        2022-01-10  False  False  True
        """
        return self._is_banned_delisted() | self._is_banned_sp()

    @staticmethod
    def rank(
        factor_df: pd.DataFrame, quantity: Optional[int] = None, ascending: bool = True
    ):
        """Compute numerical data ranks (1 through quantity) along axis.

        Parameters
        ----------
        factor_df : pd.DataFrame
            Dataframe of numerical data.
        quantity : int, optional, default None
            Number of ranks to compute. Default is None, which means all ranks.
        ascending : bool, default True
            Whether or not the elements should be ranked in ascending order.

        Returns
        -------
        pd.DataFrame with data ranks as values.

        Examples
        --------
        >>> from ezyquant import SETSignalCreator
        >>> df = pd.DataFrame(
        ...     [
        ...         [11.0, 12.0, 13.0],
        ...         [21.0, float("nan"), 23.0],
        ...         [31.0, 31.0, 31.0],
        ...     ]
        ... )
        >>> SETSignalCreator.rank(df)
             0    1    2
        0  1.0  2.0  3.0
        1  1.0  NaN  2.0
        2  1.0  1.0  1.0
        """
        df = factor_df.rank(ascending=ascending, axis=1, method="min")
        if quantity != None:
            if quantity < 1:
                raise InputError(
                    f"quantity must be greater than 0. but {quantity} is given."
                )
            df = df.mask(df > quantity, np.nan)
        return df

    """
    Protected methods
    """

    @lru_cache
    def _get_trading_dates(self) -> List[str]:
        end = self._end_date
        if self._end_date == None:
            end = self._sdr.last_table_update("DAILY_STOCK_TRADE")
        return self._sdr.get_trading_dates(start_date=self._start_date, end_date=end)

    @lru_cache
    def _get_symbol_in_universe(self) -> List[str]:
        out = set()
        if self._symbol_list:
            df = self._sdr.get_symbol_info(
                symbol_list=self._symbol_list, sec_type="S", native="L"
            )
            out.update(df["symbol"])
        if fld.MARKET_SET in self._index_list:
            df = self._sdr.get_symbol_info(
                market=fld.MARKET_SET, sec_type="S", native="L"
            )
            out.update(df["symbol"])
        if fld.MARKET_MAI.upper() in self._index_list:
            df = self._sdr.get_symbol_info(
                market=fld.MARKET_MAI, sec_type="S", native="L"
            )
            out.update(df["symbol"])

        index_list = [i for i in self._index_list if i not in (fld.MARKET_MAP_UPPER)]
        for i in index_list:
            df = self._get_symbols_by_index(i)
            out.update(df["symbol"])

        return sorted(list(out))

    def _get_symbol_info(self) -> pd.DataFrame:
        s = self._get_symbol_in_universe()
        return self._sdr.get_symbol_info(symbol_list=s, sec_type="S", native="L")

    def _reindex_trade_date(self, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
        td = self._get_trading_dates()
        return self._reindex_date(df=df, index=td, **kwargs)

    def _reindex_columns_symbol(
        self, df: pd.DataFrame, fill_value=None
    ) -> pd.DataFrame:
        s = self._get_symbol_in_universe()
        return df.reindex(columns=s, fill_value=fill_value)  # type: ignore

    def _rolling(self, data, method: str, period: int):
        roll = data.rolling(period)
        try:
            data = getattr(roll, method)()
        except AttributeError:
            raise InputError(
                f"{method} is invalid method. Please read document to check valid method."
            )
        return data

    def _rolling_skip_na_keep_inf(
        self, series: pd.Series, method: str, period: int, shift: int
    ) -> pd.Series:
        series = series.dropna().shift(shift)

        if method != fld.METHOD_CONSTANT:
            is_inf = np.isinf(series)
            series = self._rolling(series, method=method, period=period)
            series = series.mask(is_inf, np.inf)

        series = series.fillna(method="ffill")

        return series

    @lru_cache
    def _get_start_as_of_security_index(self) -> Dict[str, str]:
        out = self._sdr._get_last_as_of_date_in_security_index(
            current_date=self._start_date
        )
        return {k.upper(): v for k, v in out.items()}

    def _get_symbols_by_index(self, index: str) -> pd.DataFrame:
        start_as_of_dict = self._get_start_as_of_security_index()

        return self._sdr.get_symbols_by_index(
            index_list=[index],
            start_date=start_as_of_dict[index],
            end_date=self._end_date,
        )

    def _is_universe_static(self, universe: str) -> pd.DataFrame:
        if universe in fld.MARKET_MAP_UPPER:
            index_type = "market"
        elif universe in fld.INDUSTRY_LIST:
            index_type = "industry"
        elif universe in fld.SECTOR_LIST:
            index_type = "sector"
        else:
            raise InputError(
                f"{universe} is invalid universe. Please read document to check valid universe."
            )

        df = self._get_symbol_info().set_index("symbol")
        tds = self._get_trading_dates()

        is_uni_dict = (df[index_type].str.upper() == universe).to_dict()
        df = pd.DataFrame(is_uni_dict, index=pd.DatetimeIndex(tds))

        # Reindex columns
        df = df.reindex(columns=sorted(df.columns))  # type: ignore

        return df

    def _is_universe_dynamic(self, universe: str) -> pd.DataFrame:
        if universe not in fld.INDEX_LIST_UPPER:
            raise InputError(
                f"{universe} is invalid universe. Please read document to check valid universe."
            )

        df = self._get_symbols_by_index(universe)

        # Pivot
        df = utils.pivot_remove_index_name(
            df=df, index="as_of_date", columns="symbol", values="index"
        )

        # To bool
        df = df.notna()

        # Reindex
        df = self._reindex_trade_date(df, method="ffill", fill_value=False)
        df = self._reindex_columns_symbol(df, fill_value=False)

        return df

    def _is_banned_delisted(self) -> pd.DataFrame:
        symbol_list = self._get_symbol_in_universe()

        df = self._sdr.get_delisted(
            symbol_list=symbol_list,
            end_date=self._end_date,
        )

        df["true"] = True
        df = utils.pivot_remove_index_name(
            df=df, index="delisted_date", columns="symbol", values="true"
        )

        # Reindex
        df = self._reindex_trade_date(df, method="ffill", fill_value=False)
        df = self._reindex_columns_symbol(df, fill_value=False)

        return df

    def _is_banned_sp(self) -> pd.DataFrame:
        symbol_list = self._get_symbol_in_universe()

        df = self._sdr.get_sign_posting(
            symbol_list=symbol_list,
            end_date=self._end_date,
            sign_list=["SP"],
        )

        # hold_date and release_date can be with time
        df["hold_date"] = df["hold_date"].dt.normalize()
        df["release_date"] = df["release_date"].dt.normalize()

        df["true"] = 1
        df["false"] = 0

        # pivot_table also drop duplicated index
        df_hold = df.pivot_table(index="hold_date", columns="symbol", values="true")
        df_release = df.pivot_table(
            index="release_date", columns="symbol", values="false"
        )

        df_hold, df_release = df_hold.align(df_release)
        assert isinstance(df_release, pd.DataFrame)
        assert isinstance(df_hold, pd.DataFrame)

        df = df_release.fillna(df_hold)
        df.index.name = None
        df.columns.name = None

        # Reindex
        df = self._reindex_trade_date(df, method="ffill", fill_value=0)
        df = self._reindex_columns_symbol(df, fill_value=0)

        df = df.astype(bool)

        return df

    """
    Static methods
    """

    @staticmethod
    def _reindex_date(
        df: pd.DataFrame, index, method: Optional[str] = None, fill_value=None
    ) -> pd.DataFrame:
        """Reindex and fillna with method and fill_value."""
        index = pd.DatetimeIndex(index)

        start = pd.Series([df.index.min(), index.min()]).min()
        end = pd.Series([df.index.max(), index.max()]).max()
        dr = pd.date_range(start=start, end=end)
        df = df.reindex(dr)  # type: ignore

        if method != None:
            df = df.fillna(method=method)  # type: ignore
        if fill_value != None:
            df = df.fillna(fill_value)

        return df.reindex(index=index)  # type: ignore



from dataclasses import dataclass

import pandas as pd


@dataclass
class SETPosition:
    symbol: str
    volume: float = 0.0
    cost_price: float = 0.0
    close_price: float = float("nan")

    def __post_init__(self):
        # symbol
        assert isinstance(
            self.symbol, str
        ), f"symbol must be str, got {type(self.symbol)}"

        # volume
        assert self.volume >= 0, f"volume must be positive, got {self.volume}"
        assert (
            self.volume % 100 == 0
        ), f"volume must be multiple of 100, got {self.volume}"

        # price
        assert (
            self.cost_price >= 0
        ), f"cost_price must be positive, got {self.cost_price}"

    @property
    def cost_value(self) -> float:
        return self.volume * self.cost_price

    @property
    def close_value(self) -> float:
        if pd.notna(self.close_price):
            return self.volume * self.close_price
        else:
            return 0.0

    def match_order(self, volume: float, price: float) -> float:
        if volume > 0:
            self.cost_price = (self.cost_value + (volume * price)) / (
                self.volume + volume
            )

        self.volume += volume

        if self.volume < 0:
            raise ValueError("Insufficient volume")

        return self.volume



from typing import Callable, List

import pandas as pd

from .. import fields as fld
from .. import utils
from ..creator import SETSignalCreator
from ..errors import InputError
from ..reader import SETBusinessDay
from ..report import SETBacktestReport
from ._backtest import _backtest
from .context import Context


def backtest(
    signal_df: pd.DataFrame,
    backtest_algorithm: Callable[[Context], float],
    start_date: str,
    end_date: str,
    initial_cash: float,
    pct_commission: float = 0.0,
    pct_buy_slip: float = 0.0,
    pct_sell_slip: float = 0.0,
    price_match_mode: str = "open",
    signal_delay_bar: int = 1,
) -> SETBacktestReport:
    """Backtest function.

    Parameters
    ----------
    signal_df : pd.DataFrame
        dataframe of signal.
        index is trade date, columns are symbol, values are signal.
        missing signal in trade date will be filled with nan.
    backtest_algorithm: Callable[[Context], float],
        function for calculate trade volume.

        Parameters:
            - context: Context
                  context for backtest

        Return:
            - trade_volume: float
                  positive for buy, negative for sell, 0 or nan for no trade
    start_date : str
        start date in format YYYY-MM-DD
    end_date : str
        end date in format YYYY-MM-DD
    initial_cash : float
        cash at the beginning of the backtest
    pct_buy_slip : float, default 0.0
        percentage of buy slip, higher value means higher buy price ex. 1.0 means 1% increase
    pct_sell_slip : float, default 0.0
        percentage of sell slip, higher value means lower sell price ex. 1.0 means 1% decrease
    pct_commission : float, default 0.0
        percentage of commission fee ex. 1.0 means 1% fee
    price_match_mode : str, by default "open"
        price match mode.
            - open
            - high
            - low
            - close
            - median - (high + low)/2
            - typical - (high + low + close)/3
            - weighted - (high + low + close + close)/4
    signal_delay_bar : int, by default 1
        delay bar for shift signal.

    Returns
    -------
    SETBacktestReport
    """
    # Price df
    before_start_date = utils.date_to_str(pd.Timestamp(start_date) - SETBusinessDay())
    symbol_list = signal_df.columns.tolist()
    price_match_df = _get_price(
        start_date=start_date,
        end_date=end_date,
        symbol_list=symbol_list,
        mode=price_match_mode,
    )
    close_price_df = _get_price(
        start_date=before_start_date,
        end_date=end_date,
        symbol_list=symbol_list,
        mode=fld.D_CLOSE,
    )

    # Signal df
    signal_df = signal_df.shift(signal_delay_bar)
    signal_df = signal_df[
        (signal_df.index >= pd.Timestamp(start_date))  # type: ignore
        & (signal_df.index <= pd.Timestamp(end_date))  # type: ignore
    ]

    # Percentage
    pct_commission /= 100
    pct_buy_slip /= 100
    pct_sell_slip /= 100

    # Backtest
    # TODO: [EZ-79] initial_position_dict
    cash_series, position_df, trade_df = _backtest(
        initial_cash=initial_cash,
        signal_df=signal_df,
        backtest_algorithm=backtest_algorithm,
        close_price_df=close_price_df,
        price_match_df=price_match_df,
        pct_buy_slip=pct_buy_slip,
        pct_sell_slip=pct_sell_slip,
        pct_commission=pct_commission,
    )

    return SETBacktestReport(
        initial_capital=initial_cash,
        pct_commission=pct_commission,
        pct_buy_slip=pct_buy_slip,
        pct_sell_slip=pct_sell_slip,
        cash_series=cash_series,
        position_df=position_df,
        trade_df=trade_df,
    )


def _get_price(
    start_date: str,
    end_date: str,
    symbol_list: List[str],
    mode: str,
) -> pd.DataFrame:
    ssc = SETSignalCreator(
        start_date=start_date,
        end_date=end_date,
        index_list=[],
        symbol_list=symbol_list,
    )

    def _get_data(field: str) -> pd.DataFrame:
        return ssc.get_data(field=field, timeframe=fld.TIMEFRAME_DAILY)

    if mode in (
        fld.PRICE_MATCH_MODE_OPEN,
        fld.PRICE_MATCH_MODE_HIGH,
        fld.PRICE_MATCH_MODE_LOW,
        fld.PRICE_MATCH_MODE_CLOSE,
    ):
        out = _get_data(mode)
    elif mode == fld.PRICE_MATCH_MODE_MEDIAN:
        h = _get_data(fld.PRICE_MATCH_MODE_HIGH)
        l = _get_data(fld.PRICE_MATCH_MODE_LOW)
        out = (h + l) / 2
    elif mode == fld.PRICE_MATCH_MODE_TYPICAL:
        h = _get_data(fld.PRICE_MATCH_MODE_HIGH)
        l = _get_data(fld.PRICE_MATCH_MODE_LOW)
        c = _get_data(fld.PRICE_MATCH_MODE_CLOSE)
        out = (h + l + c) / 3
    elif mode == fld.PRICE_MATCH_MODE_WEIGHTED:
        h = _get_data(fld.PRICE_MATCH_MODE_HIGH)
        l = _get_data(fld.PRICE_MATCH_MODE_LOW)
        c = _get_data(fld.PRICE_MATCH_MODE_CLOSE)
        out = (h + l + c + c) / 4
    else:
        raise InputError(f"Invalid price_match_mode: {mode}")

    return out



import calendar
import copy
from datetime import date, datetime, timedelta
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional

import pandas as pd

from .errors import InputError


def date_to_str(value: date) -> str:
    return value.strftime("%Y-%m-%d")


def str_to_datetime(value: str) -> datetime:
    return datetime.strptime(value, "%Y-%m-%d")


def str_to_date(value: str) -> date:
    return str_to_datetime(value).date()


def str_date_add_timedelta(value: str, delta: timedelta) -> str:
    return date_to_str((str_to_date(value) + delta))


def pivot_remove_index_name(df: pd.DataFrame, *args, **kwargs) -> pd.DataFrame:
    df = df.pivot(*args, **kwargs)
    df.index.name = None
    df.columns.name = None
    return df


def round_down(value, base=100.0):
    return int(value / base) * base


def is_rebalance_weekly(
    trade_date_index: pd.DatetimeIndex, rebalance_at: int
) -> pd.Series:
    """Return a series of bool, True if the date is a rebalance date. Always
    rebalance on first day of index.

    Parameters
    ----------
    trade_date_index : pd.DatetimeIndex
        Trade date index.
    rebalance_at : int
        Day of week to rebalance. 1 is Monday, 5 is Friday.

    Returns
    -------
    pd.Series
        Series of bool, True if the date is a rebalance date.
    """
    if not trade_date_index.is_monotonic or not trade_date_index.is_unique:
        raise InputError("trade_date_index must be monotonic and unique")
    if rebalance_at < 1 or rebalance_at > 5:
        raise InputError("rebalance_at must be between 1 and 5")

    td_s = trade_date_index.to_series()

    rule = f"W-{calendar.day_abbr[rebalance_at-2]}"
    rbs = td_s.resample(rule=rule).first()
    return td_s.isin(rbs)


def is_rebalance_monthly(
    trade_date_index: pd.DatetimeIndex, rebalance_at: int
) -> pd.Series:
    """Return a series of bool, True if the date is a rebalance date. Always
    rebalance on first day of index.

    Parameters
    ----------
    trade_date_index : pd.DatetimeIndex
        Trade date index.
    rebalance_at : int
        Day of month to rebalance. Can be 1-31.

    Returns
    -------
    pd.Series
        Series of bool, True if the date is a rebalance date.
    """
    if not trade_date_index.is_monotonic or not trade_date_index.is_unique:
        raise InputError("trade_date_index must be monotonic and unique")
    if rebalance_at < 1 or rebalance_at > 31:
        raise InputError("rebalance_at must be between 1 and 31")

    td_s = trade_date_index.to_series()

    # TODO: improve is_rebalance_monthly
    v = _date_range(
        start_month=td_s[0].month,
        start_year=td_s[0].year,
        end_month=td_s[-1].month,
        end_year=td_s[-1].year,
        day_of_month=rebalance_at,
    )

    idx = td_s.searchsorted(v, side="left")
    idx = idx[idx != td_s.size]

    return td_s.isin(td_s[idx])


def _date_range(
    start_month: int,
    start_year: int,
    end_month: int,
    end_year: int,
    day_of_month: int,
):
    out = []

    ym_start = 12 * start_year + start_month - 2
    ym_end = 12 * end_year + end_month + 1

    for ym in range(ym_start, ym_end):
        y, m = divmod(ym, 12)
        try:
            out.append(date(y, m + 1, day_of_month))
        except ValueError:
            assert day_of_month > 28
            out.append(date(y, m + 2, 1))

    assert len(out) >= 3

    out = pd.to_datetime(out)
    return out


def count_true_consecutive(s: pd.Series) -> int:
    """Count the number of consecutive True values in a series."""
    out = (s * (s.groupby((s != s.shift()).cumsum()).cumcount() + 1)).max()
    if pd.isna(out):
        return 0
    else:
        return out


"""
Cache
"""


def cache_wrapper(method):
    """Cache the result of a method using lru_cache.

    Prase list arguments to sorted tuple and return copy of result.
    """
    method = lru_cache(maxsize=128)(method)

    def _arg_handler(arg):
        if isinstance(arg, list):
            return tuple(sorted(arg))
        else:
            return arg

    def wrapped(*args, **kwargs):
        new_args = tuple(_arg_handler(i) for i in args)
        new_kwargs = {k: _arg_handler(v) for k, v in kwargs.items()}
        out = method(*new_args, **new_kwargs)
        return copy.deepcopy(out)

    return wrapped


class CacheMetaClass(type):
    """Meta class for cache_wrapper."""

    def __new__(cls, clsname, bases, attrs):
        new_attrs = {
            k: cache_wrapper(v) if not k.startswith("_") and callable(v) else v
            for k, v in attrs.items()
        }
        return type.__new__(cls, clsname, bases, new_attrs)


def wrap_cache_class(cls):
    return CacheMetaClass(cls.__name__, cls.__bases__, cls.__dict__)


def cache_dataframe_wrapper(method: Callable):
    """Wrap a method to cache the result of the method.

    Parameters
    ----------
    method : Callable
        method parameter must be
            - field: str
            - symbol_list: Optional[List[str]] = None
            - start_date: Optional[str] = None
            - end_date: Optional[str] = None

        method must return dataframe with timestamp as index and symbol as column.
    """

    call_dict: Dict[str, Dict[str, Any]] = {}

    def wrapped(
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ):
        if symbol_list != None:
            symbol_list = [i.upper() for i in symbol_list]

        if field not in call_dict:
            c_symbol_list = symbol_list
            c_start_date = start_date
            c_end_date = end_date
        else:
            c_symbol_list = call_dict[field]["symbol_list"]
            c_start_date = call_dict[field]["start_date"]
            c_end_date = call_dict[field]["end_date"]

            if c_symbol_list == None or symbol_list == None:
                c_symbol_list = None
            elif set(symbol_list) - set(c_symbol_list):
                c_symbol_list = c_symbol_list + list(
                    set(symbol_list) - set(c_symbol_list)
                )

            if c_start_date == None or start_date == None:
                c_start_date = None
            elif c_start_date > start_date:
                c_start_date = start_date

            if c_end_date == None or end_date == None:
                c_end_date = None
            elif c_end_date < end_date:
                c_end_date = end_date

        df = method(
            field=field,
            symbol_list=c_symbol_list,
            start_date=c_start_date,
            end_date=c_end_date,
        )
        call_dict[field] = {
            "symbol_list": c_symbol_list,
            "start_date": c_start_date,
            "end_date": c_end_date,
        }

        if symbol_list != None:
            df = df[[i for i in symbol_list if i in df.columns]]
        if start_date != None:
            df = df[df.index >= start_date]
        if end_date != None:
            df = df[df.index <= end_date]

        return df.copy()

    return wrapped



from dataclasses import dataclass
from typing import Any

import pandas as pd

nan = float("nan")


@dataclass
class Context:
    """
    Context class for backtest.
    """

    ts: pd.Timestamp
    symbol: str = ""
    signal: Any = nan
    close_price: float = nan
    volume: float = nan
    cost_price: float = nan

    cash: float = nan
    total_cost_value: float = nan
    total_market_value: float = nan
    port_value: float = nan

    def buy_pct_port(self, pct_port: float) -> float:
        """Calculate buy volume from percentage of SETAccount. Using last close
        price. Symbol must be selected.

        Parameters
        ----------
        pct_port : float
            percentage of SETAccount

        Returns
        -------
        float
            buy volume, always positive, not round 100
        """
        return self.buy_value(self.port_value * pct_port)

    def buy_value(self, value: float) -> float:
        """Calculate buy volume from value. Using last close price. Symbol must
        be selected.

        Parameters
        ----------
        value : float
            value

        Returns
        -------
        float
            buy volume, always positive, not round 100
        """
        return value / self.close_price

    def buy_pct_position(self, pct_position: float) -> float:
        """Calculate buy volume from percentage of current position. Symbol
        must be selected.

        Parameters
        ----------
        pct_position : float
            percentage of position

        Returns
        -------
        float
            buy volume, always positive, not round 100
        """
        return pct_position * self.volume

    def sell_pct_port(self, pct_port: float) -> float:
        """Calculate sell volume from percentage of SETAccount. Using last
        close price. Symbol must be selected.

        Parameters
        ----------
        pct_port : float
            percentage of SETAccount

        Returns
        -------
        float
            sell volume, always negative, not round 100
        """
        return self.buy_pct_port(-pct_port)

    def sell_value(self, value: float) -> float:
        """Calculate sell volume from value. Using last close price. Symbol
        must be selected.

        Parameters
        ----------
        value : float
            value

        Returns
        -------
        float
            sell volume, always negative, not round 100
        """
        return self.buy_value(-value)

    def sell_pct_position(self, pct_position: float) -> float:
        """Calculate sell volume from percentage of current position. Symbol
        must be selected.

        Parameters
        ----------
        pct_position : float
            percentage of position

        Returns
        -------
        float
            sell volume, always negative, not round 100
        """
        return self.buy_pct_position(-pct_position)

    def target_pct_port(self, pct_port: float) -> float:
        """Calculate buy/sell volume to make position reach percentage of
        SETAccount. Using last close price. Symbol must be selected.

        Parameters
        ----------
        pct_port : float
            percentage of SETAccount

        Returns
        -------
        float
            buy/sell volume, not round 100
        """
        return self.buy_pct_port(pct_port) - self.volume

    def target_value(self, value: float) -> float:
        """Calculate buy/sell volume to make position reach value. Symbol must
        be selected.

        Parameters
        ----------
        value : float
            value

        Returns
        -------
        float
            buy/sell volume, not round 100
        """
        return self.buy_value(value) - self.volume



from dataclasses import fields
from typing import Callable, Dict, List, Tuple

import pandas as pd
from pandas.testing import assert_index_equal

from .. import validators as vld
from .account import SETAccount
from .context import Context
from .position import SETPosition
from .trade import SETTrade

position_df_columns = ["timestamp"] + [i.name for i in fields(SETPosition)]
trade_df_columns = [i.name for i in fields(SETTrade)]


def _backtest(
    initial_cash: float,
    signal_df: pd.DataFrame,
    backtest_algorithm: Callable[[Context], float],
    close_price_df: pd.DataFrame,
    price_match_df: pd.DataFrame,
    pct_buy_slip: float,
    pct_sell_slip: float,
    pct_commission: float,
) -> Tuple[pd.Series, pd.DataFrame, pd.DataFrame]:
    """Backtest function without load any data.

    Parameters
    ----------
    initial_cash : float
        cash at the beginning of the backtest
    signal_df : pd.DataFrame
        dataframe of signal.
        index is trade date, columns are symbol, values are signal.
        missing signal in trade date will be filled with nan.
    backtest_algorithm: Callable[[Context], float],
        function for calculate trade volume.
        Parameters:
            - context: Context
                context for backtest
        Return:
            - trade_volume: float
                positive for buy, negative for sell, 0 or nan for no trade
    close_price_df : pd.DataFrame
        dataframe of close price.
        index is trade date, columns are symbol, values are weight.
        first row will be used as initial close price.
    price_match_df : pd.DataFrame
        dataframe of match price.
        index is trade date, columns are symbol, values are weight.
        index and columns must be same as or more than close_price_df.
    pct_buy_slip : float, default 0.0
        percentage of buy slip, higher value means higher buy price ex. 0.01 means 1% increase
    pct_sell_slip : float, default 0.0
        percentage of sell slip, higher value means lower sell price ex. 0.01 means 1% decrease
    pct_commission : float, default 0.0
        percentage of commission fee ex. 0.01 means 1% fee

    Returns
    -------
    Tuple[pd.Series, pd.DataFrame, pd.DataFrame]
        Return snapshot at end of day:
            - cash_series
                - timestamp (index)
                - cash
            - position_df
                - timestamp
                - symbol
                - volume
                - cost_price
                - close_price
            - trade_df
                - matched_at
                - symbol
                - volume
                - price
                - pct_commission
    """
    vld.check_cash(initial_cash)
    vld.check_pct(pct_buy_slip)
    vld.check_pct(pct_sell_slip)
    vld.check_pct(pct_commission)

    # Ratio
    ratio_buy_slip = 1.0 + pct_buy_slip
    ratio_sell_slip = 1.0 - pct_sell_slip

    # Account
    acct = SETAccount(
        cash=initial_cash,
        pct_commission=pct_commission,
        position_dict={},  # TODO: [EZ-79] initial position dict
        trade_list=[],  # TODO: initial trade
    )
    acct.set_position_close_price(close_price_df.iloc[0].to_dict())

    # remove first close row
    close_price_df = close_price_df.iloc[1:]  # type: ignore

    # reindex signal
    signal_df = signal_df.reindex(close_price_df.index)  # type: ignore

    # Check after remove first close row
    _check_df_input(signal_df, close_price_df, price_match_df)

    # Dict
    signal_dict: Dict[pd.Timestamp, Dict[str, float]] = signal_df.to_dict("index")  # type: ignore
    close_price_dict: Dict[pd.Timestamp, Dict[str, float]] = close_price_df.to_dict(
        "index"
    )  # type: ignore
    price_match_dict: Dict[pd.Timestamp, Dict[str, float]] = price_match_df.to_dict(
        "index"
    )  # type: ignore

    position_df_list: List[pd.DataFrame] = [
        # First dataframe for sort columns
        pd.DataFrame(columns=position_df_columns, dtype="float64"),
    ]

    def calculate_trade_volume(
        ts: pd.Timestamp,
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        signal_d = signal_dict[ts]

        buy_volume_d = dict()
        sell_volume_d = dict()

        ctx = Context(
            ts=ts,
            cash=acct.cash,
            total_cost_value=acct.total_cost_value,
            total_market_value=acct.total_market_value,
            port_value=acct.port_value,
        )

        for k, v in signal_d.items():
            # Setup ctx
            ctx.symbol = k
            ctx.signal = v
            ctx.close_price = acct.close_price_dict[k]
            if k in acct.position_dict:
                pos = acct.position_dict[k]
                ctx.volume = pos.volume
                ctx.cost_price = pos.cost_price
            else:
                ctx.volume = 0.0
                ctx.cost_price = 0.0

            trade_volume = backtest_algorithm(ctx)

            if trade_volume > 0:
                buy_volume_d[k] = trade_volume
            elif trade_volume < 0:
                sell_volume_d[k] = trade_volume

        return buy_volume_d, sell_volume_d

    def on_interval(ts: pd.Timestamp, close_price_dict: Dict[str, float]) -> float:
        buy_volume_d, sell_volume_d = calculate_trade_volume(ts)

        # Trade
        for k, v in sell_volume_d.items():
            price = price_match_dict[ts][k] * ratio_sell_slip
            acct.match_order_if_possible(ts, symbol=k, volume=v, price=price)
        for k, v in buy_volume_d.items():
            price = price_match_dict[ts][k] * ratio_buy_slip
            acct.match_order_if_possible(ts, symbol=k, volume=v, price=price)

        # Snap
        acct.set_position_close_price(close_price_dict)
        pos_df = acct.position_df
        pos_df["timestamp"] = ts
        position_df_list.append(pos_df)

        return acct.cash

    cash_s = pd.Series(
        [on_interval(k, v) for k, v in close_price_dict.items()],
        index=close_price_df.index,
    )
    assert isinstance(cash_s, pd.Series)

    position_df = pd.concat(position_df_list, ignore_index=True)
    position_df["timestamp"] = pd.to_datetime(position_df["timestamp"])

    trade_df = pd.DataFrame(acct.trade_list, columns=trade_df_columns)
    trade_df["matched_at"] = pd.to_datetime(trade_df["matched_at"])

    return cash_s, position_df, trade_df


def _check_df_input(
    signal_df: pd.DataFrame,
    close_price_df: pd.DataFrame,
    price_match_df: pd.DataFrame,
):
    vld.check_df_symbol_daily(signal_df)
    vld.check_df_symbol_daily(close_price_df)
    vld.check_df_symbol_daily(price_match_df)

    assert_index_equal(close_price_df.index, price_match_df.index)
    assert_index_equal(close_price_df.columns, price_match_df.columns)

    assert_index_equal(close_price_df.index, signal_df.index)
    assert_index_equal(close_price_df.columns, signal_df.columns)



import os.path
import warnings
from datetime import date
from functools import lru_cache
from typing import Dict, List, Optional

import pandas as pd
import sqlalchemy as sa
from pandas.errors import PerformanceWarning
from pandas.tseries.offsets import CustomBusinessDay
from sqlalchemy import Column, MetaData, Table, and_, case, func, select
from sqlalchemy.exc import DatabaseError
from sqlalchemy.sql import Select
from sqlalchemy.sql.selectable import Select

from . import fields as fld
from . import utils
from . import validators as vld
from .errors import InputError

warnings.filterwarnings("ignore", category=PerformanceWarning)

TRADE_DATE = "trade_date"
NAME = "name"
VALUE = "value"


class SETDataReader:

    _sqlite_path: Optional[str] = None

    def __init__(self):
        """SETDataReader read PSIMS data."""
        if self._sqlite_path == None:
            raise InputError(
                "You need to connect sqlite using ezyquant.connect_sqlite(sqlite_path)."
            )

        self._engine = sa.create_engine(f"sqlite:///{self._sqlite_path}")
        self._metadata = MetaData(self._engine)

        if not os.path.isfile(self._sqlite_path):
            raise InputError(f"{self._sqlite_path} is not found")
        try:
            self._table("SECURITY")
        except DatabaseError as e:
            raise InputError(e)

    def last_table_update(self, table_name: str) -> str:
        """Last D_TRADE in table.

        Parameters
        ----------
        table_name : str
            name of table:
            - DAILY_STOCK_TRADE
            - DAILY_STOCK_STAT
            - MKTSTAT_DAILY_INDEX
            - MKTSTAT_DAILY_MARKET
            - DAILY_SECTOR_INFO


        Returns
        -------
        str
            string with format YYYY-MM-DD
        """
        t = self._table(table_name)
        stmt = select([func.max(func.DATE(t.c.D_TRADE))])
        res = self._execute(stmt).scalar()
        assert isinstance(res, str)
        return res

    def last_update(self) -> str:
        """Last database update, checking from last D_TRADE in following
        tables:

        - DAILY_STOCK_TRADE
        - DAILY_STOCK_STAT
        - MKTSTAT_DAILY_INDEX
        - MKTSTAT_DAILY_MARKET
        - DAILY_SECTOR_INFO

        Returns
        -------
        str
            string with format YYYY-MM-DD
        """
        d1 = self.last_table_update("DAILY_STOCK_TRADE")
        d2 = self.last_table_update("DAILY_STOCK_STAT")
        d3 = self.last_table_update("MKTSTAT_DAILY_INDEX")
        d4 = self.last_table_update("MKTSTAT_DAILY_MARKET")
        d5 = self.last_table_update("DAILY_SECTOR_INFO")
        assert d1 == d2 == d3 == d4 == d5, "database is not consistent"
        return d1

    def get_trading_dates(
        self, start_date: Optional[str] = None, end_date: Optional[str] = None
    ) -> List[str]:
        """Data from table CALENDAR.

        Parameters
        ----------
        start_date : Optional[str]
            start of D_TRADE, by default None
        end_date : Optional[str]
            end of D_TRADE, by default None

        Returns
        -------
        List[str]
            list of string with format YYYY-MM-DD
        """
        calendar_t = self._table("CALENDAR")

        stmt = select([func.DATE(calendar_t.c.D_TRADE)]).order_by(
            func.DATE(calendar_t.c.D_TRADE)
        )

        stmt = self._filter_stmt_by_date(
            stmt=stmt,
            column=calendar_t.c.D_TRADE,
            start_date=start_date,
            end_date=end_date,
        )

        res = self._execute(stmt).all()

        return [i[0] for i in res]

    def is_trading_date(self, check_date: str) -> bool:
        """Data from table CALENDAR.

        Parameters
        ----------
        check_date : str
            string with format YYYY-MM-DD

        Returns
        -------
        bool
            is trading date
        """
        calendar_t = self._table("CALENDAR")

        stmt = select([func.count(calendar_t.c.D_TRADE)]).where(
            func.DATE(calendar_t.c.D_TRADE) == check_date
        )

        res = self._execute(stmt).scalar()

        assert isinstance(res, int)
        return res > 0

    def is_today_trading_date(self) -> bool:
        """Data from table CALENDAR.

        Returns
        -------
        bool
            is today trading date
        """
        return self.is_trading_date(utils.date_to_str(date.today()))

    def get_symbol_info(
        self,
        symbol_list: Optional[List[str]] = None,
        market: Optional[str] = None,
        industry: Optional[str] = None,
        sector: Optional[str] = None,
        sec_type: Optional[str] = None,
        native: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table SECURITY.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        market : Optional[str]
            I_MARKET e.g. 'SET', 'mai', by default None
        industry : Optional[str]
            SECTOR.N_INDUSTRY, by default None
        sector : Optional[str]
            SECTOR.N_SECTOR, by default None
        sec_type : Optional[str]
            Security type, by default None
                - S: Common
                - F: Common Foreign
                - P: Preferred
                - Q: Preferred Foreign
                - U: Unit Trusts
                - T: Unit Trusts Foreign
                - W: Warrant
                - X: Depository Receipts
                - Y: Depository Receipts Foreign
                - L: ETFs
                - O: Index Options
                - D: Debenture
                - C: Convertible Debenture
                - I: Index Warrant
                - V: Derivatives Warrant
                - R: Transferable Subscription Right
                - G: Government Bond
                - J: Convertible Preferred
        native : Optional[str]
            I_NATIVE, by default None
                - L: Local
                - F: Foreign
                - U: Thai Trust Fund
                - R: NVDR

        Returns
        -------
        pd.DataFrame
            symbol info dataframe contain columns:
                - symbol_id: int - I_SECURITY
                - symbol: str - N_SECURITY
                - market: str - I_MARKET
                - industry: str - SECTOR.N_INDUSTRY
                - sector: str - SECTOR.N_SECTOR
                - sec_type: str - I_SEC_TYPE
                - native: str - I_NATIVE

        Examples
        --------
        >>> from ezyquant import fields as fld
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_symbol_info(["BBL"])
           symbol_id symbol market industry sector sec_type native
        0          1    BBL    SET  FINCIAL   BANK        S      L
        >>> sdr.get_symbol_info(sector=fld.SECTOR_MINE)
           symbol_id symbol market industry sector sec_type native
        0        167    THL    SET  RESOURC   MINE        S      L
        1        168  THL-R    SET  RESOURC   MINE        S      R
        2        169  THL-U    SET  RESOURC   MINE        S      U
        3       2968  THL-F    SET  RESOURC   MINE        F      F
        """
        security_t = self._table("SECURITY")
        sector_t = self._table("SECTOR")

        j = self._join_sector_table(security_t, isouter=True)
        stmt = (
            select(
                [
                    security_t.c.I_SECURITY.label("symbol_id"),
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    security_t.c.I_MARKET.label("market"),
                    func.trim(sector_t.c.N_INDUSTRY).label("industry"),
                    func.trim(sector_t.c.N_SECTOR).label("sector"),
                    security_t.c.I_SEC_TYPE.label("sec_type"),
                    security_t.c.I_NATIVE.label("native"),
                ]
            )
            .select_from(j)
            .order_by(security_t.c.I_SECURITY)
        )

        stmt = self._filter_str_in_list(
            stmt=stmt, column=security_t.c.N_SECURITY, values=symbol_list
        )
        if market != None:
            market = market.upper()
            stmt = stmt.where(security_t.c.I_MARKET == fld.MARKET_MAP_UPPER[market])
        if industry != None:
            industry = industry.upper()
            stmt = stmt.where(func.trim(sector_t.c.N_INDUSTRY) == industry)
        if sector != None:
            sector = sector.upper()
            stmt = stmt.where(func.trim(sector_t.c.N_SECTOR) == sector)
        if sec_type != None:
            sec_type = sec_type.upper()
            stmt = stmt.where(security_t.c.I_SEC_TYPE == sec_type)
        if native != None:
            native = native.upper()
            stmt = stmt.where(security_t.c.I_NATIVE == native)

        df = self._read_sql_query(stmt)

        map_market = {v: k for k, v in fld.MARKET_MAP.items()}
        df["market"] = df["market"].replace(map_market)

        return df

    def get_company_info(self, symbol_list: Optional[List[str]] = None) -> pd.DataFrame:
        """Data from table COMPANY.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            SECURITY.N_SECURITY in symbol_list, must be unique, by default None

        Returns
        -------
        pd.DataFrame
            company info dataframe contain columns:
                - company_id: int - I_COMPANY
                - symbol: str - SECURITY.N_SECURITY
                - company_name_t: str - N_COMPANY_T
                - company_name_e: str - N_COMPANY_E
                - address_t: str - A_COMPANY_T
                - address_e: str - A_COMPANY_E
                - zip: str - I_ZIP
                - tel: str - E_TEL
                - fax: str - E_FAX
                - email: str - E_EMAIL
                - url: str - E_URL
                - establish: date - D_ESTABLISH
                - dvd_policy_t: str - E_DVD_POLICY_T
                - dvd_policy_e: str - E_DVD_POLICY_E

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_company_info(symbol_list=["BBL", "PTT"])
           company_id symbol               company_name_t  ...  establish                                       dvd_policy_t                                       dvd_policy_e
        0           1    BBL  ธนาคารกรุงเทพ จำกัด (มหาชน)  ...  1/12/1944  เมื่อผลประกอบการของธนาคารมีกำไร (โดยมีเงื่อนไข...  Pays when company has profit (with additional ...
        1         646    PTT    บริษัท ปตท. จำกัด (มหาชน)  ...  1/10/2001  ไม่ต่ำกว่าร้อยละ 25 ของกำไรสุทธิที่เหลือหลังหั...  Not less than 25% of net income after deductio...
        """
        company_t = self._table("COMPANY")
        security_t = self._table("SECURITY")

        j = security_t.join(company_t, company_t.c.I_COMPANY == security_t.c.I_COMPANY)
        stmt = (
            select(
                [
                    company_t.c.I_COMPANY.label("company_id"),
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    company_t.c.N_COMPANY_T.label("company_name_t"),
                    company_t.c.N_COMPANY_E.label("company_name_e"),
                    company_t.c.A_COMPANY_T.label("address_t"),
                    company_t.c.A_COMPANY_E.label("address_e"),
                    company_t.c.I_ZIP.label("zip"),
                    company_t.c.E_TEL.label("tel"),
                    company_t.c.E_FAX.label("fax"),
                    company_t.c.E_EMAIL.label("email"),
                    company_t.c.E_URL.label("url"),
                    func.trim(company_t.c.D_ESTABLISH).label("establish"),
                    company_t.c.E_DVD_POLICY_T.label("dvd_policy_t"),
                    company_t.c.E_DVD_POLICY_E.label("dvd_policy_e"),
                ]
            )
            .select_from(j)
            .order_by(company_t.c.I_COMPANY)
        )

        stmt = self._filter_str_in_list(
            stmt=stmt, column=security_t.c.N_SECURITY, values=symbol_list
        )

        df = self._read_sql_query(stmt)
        return df

    def get_change_name(
        self,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table CHANGE_NAME_SECURITY.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of effect_date (D_EFFECT), by default None
        end_date : Optional[str]
            end of effect_date (D_EFFECT), by default None

        Returns
        -------
        pd.DataFrame
            change name dataframe contain columns:
                - symbol_id: int - I_SECURITY
                - symbol: str - SECURITY.N_SECURITY
                - effect_date: date - D_EFFECT
                - symbol_old: str - N_SECURITY_OLD
                - symbol_new: str - N_SECURITY_NEW

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_change_name(["SMG"])
           symbol_id symbol effect_date symbol_old symbol_new
        0        220    SMG  2006-07-31        SMG      SCSMG
        1        220    SMG  2014-08-28      SCSMG        SMG
        >>> sdr.get_change_name(start_date="2014-08-28", end_date="2014-08-29")
           symbol_id    symbol effect_date  symbol_old symbol_new
        0        220       SMG  2014-08-28       SCSMG        SMG
        1        221     SMG-F  2014-08-28     SCSMG-F      SMG-F
        2        222     SMG-R  2014-08-28     SCSMG-R      SMG-R
        3       2793    SMG-W1  2014-08-28    SCSMG-W1     SMG-W1
        4       2794    SMG-WB  2014-08-28    SCSMG-WB     SMG-WB
        5       3375  SMG-W1-R  2014-08-28  SCSMG-W1-R   SMG-W1-R
        """
        security_t = self._table("SECURITY")
        change_name_t = self._table("CHANGE_NAME_SECURITY")

        j = change_name_t.join(
            security_t, security_t.c.I_SECURITY == change_name_t.c.I_SECURITY
        )
        stmt = (
            select(
                [
                    change_name_t.c.I_SECURITY.label("symbol_id"),
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    change_name_t.c.D_EFFECT.label("effect_date"),
                    func.trim(change_name_t.c.N_SECURITY_OLD).label("symbol_old"),
                    func.trim(change_name_t.c.N_SECURITY_NEW).label("symbol_new"),
                ]
            )
            .select_from(j)
            .where(change_name_t.c.D_EFFECT != None)
            .where(
                func.trim(change_name_t.c.N_SECURITY_OLD)
                != func.trim(change_name_t.c.N_SECURITY_NEW)
            )
            .order_by(func.DATE(change_name_t.c.D_EFFECT))
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=change_name_t.c.D_EFFECT,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt)
        df = df.drop_duplicates(ignore_index=True)
        return df

    def get_dividend(
        self,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        ca_type_list: Optional[List[str]] = None,
        adjusted_list: List[str] = ["", "CR", "PC", "RC", "SD", "XR"],
    ) -> pd.DataFrame:
        """Data from table RIGHTS_BENEFIT. Include only Cash Dividend (CD) and
        Stock Dividend (SD). Not include Cancelled (F_CANCEL!='C') and dps more
        than 0 (Z_RIGHTS>0). ex_date and pay_date can be null.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of ex_date (D_SIGN), by default None
        end_date : Optional[str]
            end of ex_date (D_SIGN), by default None
        ca_type_list : Optional[List[str]]
            Coperatie action type (N_CA_TYPE), by default None
                - CD - cash dividend
                - SD - stock dividend
        adjusted_list : List[str]
            Adjust data by ca_type, empty list for no adjust, by default ["", "CR", "PC", "RC", "SD", "XR"]

        Returns
        -------
        pd.DataFrame
            dividend dataframe contain columns:
                - symbol: str - SECURITY.N_SECURITY
                - ex_date: date - D_SIGN
                - pay_date: date - D_BEG_PAID
                - ca_type: str - N_CA_TYPE
                - dps: int - Z_RIGHTS

        Notes
        -----
        - ex_date and pay_date can be null.
        - pay_date can be non trade date.

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_dividend(["M"])
           symbol     ex_date    pay_date ca_type  dps
        0       M  2014-05-06  2014-05-21      CD  1.6
        1       M  2014-08-20  2014-09-04      CD  0.8
        2       M  2015-04-30  2015-05-21      CD  1.0
        3       M  2015-08-24  2015-09-08      CD  0.9
        4       M  2016-04-28  2016-05-19      CD  1.0
        5       M  2016-08-23  2016-09-08      CD  1.0
        6       M  2017-05-04  2017-05-23      CD  1.1
        7       M  2017-08-22  2017-09-08      CD  1.1
        8       M  2018-05-07  2018-05-23      CD  1.2
        9       M  2018-08-22  2018-09-07      CD  1.2
        10      M  2019-05-07  2019-05-23      CD  1.3
        11      M  2019-08-26  2019-09-11      CD  1.3
        12      M  2020-04-22  2020-05-08      CD  1.3
        13      M  2020-08-24  2020-09-10      CD  0.5
        14      M  2021-05-10  2021-05-25      CD  0.5
        15      M  2022-05-10  2022-05-25      CD  0.8
        >>> sdr.get_dividend(["M"], start_date="2020-07-28", end_date="2022-05-11")
            symbol     ex_date    pay_date ca_type  dps
        0        M  2020-08-24  2020-09-10      CD  0.5
        1        M  2021-05-10  2021-05-25      CD  0.5
        2        M  2022-05-10  2022-05-25      CD  0.8
        """
        security_t = self._table("SECURITY")
        rights_benefit_t = self._table("RIGHTS_BENEFIT")

        j = rights_benefit_t.join(
            security_t, security_t.c.I_SECURITY == rights_benefit_t.c.I_SECURITY
        )
        stmt = (
            select(
                [
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    rights_benefit_t.c.D_SIGN.label("ex_date"),
                    rights_benefit_t.c.D_BEG_PAID.label("pay_date"),
                    func.trim(rights_benefit_t.c.N_CA_TYPE).label("ca_type"),
                    rights_benefit_t.c.Z_RIGHTS.label("dps"),
                ]
            )
            .select_from(j)
            .where(func.trim(rights_benefit_t.c.N_CA_TYPE).in_(["CD", "SD"]))
            .where(func.trim(rights_benefit_t.c.F_CANCEL) != "C")
            .where(rights_benefit_t.c.Z_RIGHTS > 0)
            .order_by(func.DATE(rights_benefit_t.c.D_SIGN))
        )

        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=rights_benefit_t.c.D_SIGN,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )
        stmt = self._filter_str_in_list(
            stmt=stmt, column=rights_benefit_t.c.N_CA_TYPE, values=ca_type_list
        )

        df = self._read_sql_query(stmt)

        df = self._merge_adjust_factor_dividend(df, adjusted_list=adjusted_list)

        return df

    def get_delisted(
        self,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table SECURITY_DETAIL. Filter delisted by D_DELISTED !=
        None.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of delisted_date (D_DELISTED), by default None
        end_date : Optional[str]
            end of delisted_date (D_DELISTED), by default None

        Returns
        -------
        pd.DataFrame
            delisted dataframe contain columns:
                - symbol: str - SECURITY.N_SECURITY
                - delisted_date: date - D_DELISTED

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_delisted(start_date="2020-02-20", end_date="2020-02-20")
             symbol delisted_date
        0    ROBINS    2020-02-20
        1    KK202A    2020-02-20
        2  CB20220A    2020-02-20
        3  CB20220B    2020-02-20
        """
        security_t = self._table("SECURITY")
        security_detail_t = self._table("SECURITY_DETAIL")

        j = security_detail_t.join(
            security_t, security_t.c.I_SECURITY == security_detail_t.c.I_SECURITY
        )
        stmt = (
            select(
                [
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    security_detail_t.c.D_DELISTED.label("delisted_date"),
                ]
            )
            .select_from(j)
            .where(security_detail_t.c.D_DELISTED != None)
            .order_by(func.DATE(security_detail_t.c.D_DELISTED))
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=security_detail_t.c.D_DELISTED,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt)
        return df

    def get_sign_posting(
        self,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        sign_list: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Data from table SIGN_POSTING.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of hold_date (D_HOLD), by default None
        end_date : Optional[str]
            end of hold_date (D_HOLD), by default None
        sign_list : Optional[List[str]]
            N_SIGN in sign_list, by default None
                - C - Caution Flag
                - CM - Call Market
                - DS - Designated
                - H - Halt
                - NC - Non Compliance
                - NP - Notice Pending
                - SP - Suspension
                - ST

        Returns
        -------
        pd.DataFrame
            sp dataframe contain columns:
                - symbol: str - SECURITY.N_SECURITY
                - hold_date: date - D_HOLD
                - release_date: date - D_RELEASE
                - sign: str - N_SIGN

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_sign_posting(symbol_list=["THAI"], start_date="2020-11-12", end_date="2021-02-25")
          symbol  hold_date sign
        0   THAI 2020-11-12   SP
        1   THAI 2021-02-25   SP
        """
        security_t = self._table("SECURITY")
        sign_posting_t = self._table("SIGN_POSTING")

        j = sign_posting_t.join(
            security_t, security_t.c.I_SECURITY == sign_posting_t.c.I_SECURITY
        )
        stmt = (
            select(
                [
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    sign_posting_t.c.D_HOLD.label("hold_date"),
                    sign_posting_t.c.D_RELEASE.label("release_date"),
                    func.trim(sign_posting_t.c.N_SIGN).label("sign"),
                ]
            )
            .select_from(j)
            .order_by(func.DATE(sign_posting_t.c.D_HOLD))
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=sign_posting_t.c.D_HOLD,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )
        stmt = self._filter_str_in_list(
            stmt=stmt, column=sign_posting_t.c.N_SIGN, values=sign_list
        )

        df = self._read_sql_query(stmt)
        return df

    def get_symbols_by_index(
        self,
        index_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table SECURITY_INDEX.

        Parameters
        ----------
        index_list : Optional[List[str]]
            index (SECTOR.N_SECTOR)
                - SETWB
                - SETTHSI
                - SETCLMV
                - SETHD
                - sSET
                - SET100
                - SET50
        start_date : Optional[str]
            start of as_of_date (D_AS_OF), by default None
        end_date : Optional[str]
            end of as_of_date (D_AS_OF), by default None

        Returns
        -------
        pd.DataFrame
            - as_of_date: date - D_AS_OF
            - index: str - SECTOR.N_SECTOR
            - symbol: str - SECURITY.N_SECURITY
            - seq: int - SECURITY_INDEX.S_SEQ

        Note
        -------
        - SET50 filter S_SEQ 1-50
        - SET100 filter S_SEQ 1-100
        - SETHD filter S_SEQ 1-30

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_symbols_by_index(index_list=["SET50"], start_date="2022-01-04", end_date="2022-01-04")
           as_of_date  index  symbol  seq
        0  2022-01-04  SET50     OSP    1
        1  2022-01-04  SET50     CBG    2
        2  2022-01-04  SET50      TU    3
        3  2022-01-04  SET50    MINT    4
        4  2022-01-04  SET50     CPF    5
        5  2022-01-04  SET50    STGT    6
        6  2022-01-04  SET50   TISCO    7
        7  2022-01-04  SET50     KTB    8
        8  2022-01-04  SET50     TTB    9
        9  2022-01-04  SET50     SCB   10
        10 2022-01-04  SET50   KBANK   11
        11 2022-01-04  SET50     BBL   12
        12 2022-01-04  SET50  TIDLOR   13
        13 2022-01-04  SET50     MTC   14
        14 2022-01-04  SET50   SAWAD   15
        15 2022-01-04  SET50     KTC   16
        16 2022-01-04  SET50   PTTGC   17
        17 2022-01-04  SET50     IVL   18
        18 2022-01-04  SET50    SCGP   19
        19 2022-01-04  SET50     SCC   20
        20 2022-01-04  SET50     AWC   21
        21 2022-01-04  SET50     CPN   22
        22 2022-01-04  SET50      LH   23
        23 2022-01-04  SET50      OR   24
        24 2022-01-04  SET50    GULF   25
        25 2022-01-04  SET50   BGRIM   26
        26 2022-01-04  SET50    GPSC   27
        27 2022-01-04  SET50      EA   28
        28 2022-01-04  SET50     TOP   29
        29 2022-01-04  SET50     PTT   30
        30 2022-01-04  SET50   RATCH   31
        31 2022-01-04  SET50    IRPC   32
        32 2022-01-04  SET50    EGCO   33
        33 2022-01-04  SET50   PTTEP   34
        34 2022-01-04  SET50   BANPU   35
        35 2022-01-04  SET50     CRC   36
        36 2022-01-04  SET50    COM7   37
        37 2022-01-04  SET50  GLOBAL   38
        38 2022-01-04  SET50   CPALL   39
        39 2022-01-04  SET50   HMPRO   40
        40 2022-01-04  SET50    BDMS   41
        41 2022-01-04  SET50      BH   42
        42 2022-01-04  SET50     BEM   43
        43 2022-01-04  SET50     AOT   44
        44 2022-01-04  SET50     BTS   45
        45 2022-01-04  SET50    DTAC   46
        46 2022-01-04  SET50    TRUE   47
        47 2022-01-04  SET50  ADVANC   48
        48 2022-01-04  SET50  INTUCH   49
        49 2022-01-04  SET50     KCE   50
        """
        security_t = self._table("SECURITY")
        sector_t = self._table("SECTOR")
        security_index_t = self._table("SECURITY_INDEX")

        j = self._join_sector_table(security_index_t).join(
            security_t, security_t.c.I_SECURITY == security_index_t.c.I_SECURITY
        )
        stmt = (
            select(
                [
                    security_index_t.c.D_AS_OF.label("as_of_date"),
                    func.trim(sector_t.c.N_SECTOR).label("index"),
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    security_index_t.c.S_SEQ.label("seq"),
                ]
            )
            .select_from(j)
            .where(
                case(
                    (
                        func.trim(sector_t.c.N_SECTOR) == fld.INDEX_SET50,
                        security_index_t.c.S_SEQ <= 50,
                    ),
                    (
                        func.trim(sector_t.c.N_SECTOR) == fld.INDEX_SET100,
                        security_index_t.c.S_SEQ <= 100,
                    ),
                    (
                        func.trim(sector_t.c.N_SECTOR) == fld.INDEX_SETHD,
                        security_index_t.c.S_SEQ <= 30,
                    ),
                    else_=True,
                )
            )
            .order_by(
                func.DATE(security_index_t.c.D_AS_OF),
                sector_t.c.N_SECTOR,
                security_index_t.c.S_SEQ,
            )
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=sector_t.c.N_SECTOR,
            date_column=security_index_t.c.D_AS_OF,
            symbol_list=index_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt)

        return df

    def get_adjust_factor(
        self,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        ca_type_list: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Data from table ADJUST_FACTOR.

        Parameters
        ----------
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of effect_date (D_EFFECT), by default None
        end_date : Optional[str]
            end of effect_date (D_EFFECT), by default None
        ca_type_list : Optional[List[str]]
            Coperatie action type (N_CA_TYPE), by default None
                - '  '
                - 'CR' - Capital Reduction
                - 'PC' - Par Change
                - 'RC' - Ratio Change
                - 'SD' - Stock Dividend
                - 'XR' - Rights

        Returns
        -------
        pd.DataFrame
            adjust factor dataframe contain columns:
                - symbol: str - SECURITY.N_SECURITY
                - effect_date: date - D_EFFECT
                - ca_type: str - N_CA_TYPE
                - adjust_factor: float - R_ADJUST_FACTOR

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> sdr = SETDataReader()
        >>> sdr.get_adjust_factor(symbol_list=["RAM"])
          symbol effect_date ca_type  adjust_factor
        0    RAM  2019-06-17      PC           0.05
        1    RAM  2021-11-09      PC           0.20
        """
        security_t = self._table("SECURITY")
        adjust_factor_t = self._table("ADJUST_FACTOR")

        j = adjust_factor_t.join(
            security_t, security_t.c.I_SECURITY == adjust_factor_t.c.I_SECURITY
        )
        stmt = (
            select(
                [
                    func.trim(security_t.c.N_SECURITY).label("symbol"),
                    adjust_factor_t.c.D_EFFECT.label("effect_date"),
                    adjust_factor_t.c.N_CA_TYPE.label("ca_type"),
                    adjust_factor_t.c.R_ADJUST_FACTOR.label("adjust_factor"),
                ]
            )
            .select_from(j)
            .order_by(func.DATE(adjust_factor_t.c.D_EFFECT))
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=adjust_factor_t.c.D_EFFECT,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )
        stmt = self._filter_str_in_list(
            stmt=stmt, column=adjust_factor_t.c.N_CA_TYPE, values=ca_type_list
        )

        df = self._read_sql_query(stmt)

        return df

    def get_data_symbol_daily(
        self,
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        adjusted_list: List[str] = ["", "CR", "PC", "RC", "SD", "XR"],
    ) -> pd.DataFrame:
        """Data from table DAILY_STOCK_TRADE, DAILY_STOCK_STAT.

        Filter only Auto Matching (I_TRADING_METHOD='A').

        Parameters
        ----------
        field : str
            - prior
            - open
            - high
            - low
            - close
            - average
            - last_bid
            - last_offer
            - trans
            - volume
            - value
            - pe
            - pb
            - par
            - dps
            - dvd_yield
            - mkt_cap
            - eps
            - book_value
            - quarter_fin
            - month_dvd
            - as_of
            - dividend
            - status
            - benefit
            - share_listed
            - turnover
            - share_index
            - npg
            - total_volume
            - total_value
            - beta
            - roi
            - acc_dps
            - dvd_payment
            - dvd_payout
            - earning
            - iv
            - delta
            - notice
            - non_compliance
            - stabilization
            - call_market
            - caution
            - 12m_dvd_yield
            - peg
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of trade_date (D_TRADE), by default None
        end_date : Optional[str]
            end of trade_date (D_TRADE), by default None
        adjusted_list : List[str]
            Adjust data by ca_type, empty list for no adjust, by default ["", "CR", "PC", "RC", "SD", "XR"]

        Returns
        -------
        pd.DataFrame
            - symbol(N_SECURITY): str as column
            - trade_date(D_TRADE): date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_symbol_daily(
        ...    field=fld.D_CLOSE,
        ...    symbol_list=["COM7", "MALEE"],
        ...    start_date="2022-01-01",
        ...    end_date="2022-01-10",
        ... )
                      COM7  MALEE
        2022-01-04  41.875   6.55
        2022-01-05  41.625   6.50
        2022-01-06  41.500   6.50
        2022-01-07  41.000   6.40
        2022-01-10  40.875   6.30
        """
        adjusted_list = list(adjusted_list)  # copy to avoid modify original list
        field = field.lower()

        if symbol_list:
            symbol_list = [i.upper() for i in symbol_list]

        security_t = self._table("SECURITY")

        if field in fld.DAILY_STOCK_TRADE_MAP:
            daily_stock_t = self._table("DAILY_STOCK_TRADE")
            value_col = daily_stock_t.c[fld.DAILY_STOCK_TRADE_MAP[field]]
        elif field in fld.DAILY_STOCK_STAT_MAP:
            daily_stock_t = self._table("DAILY_STOCK_STAT")
            value_col = daily_stock_t.c[fld.DAILY_STOCK_STAT_MAP[field]]
        else:
            raise InputError(
                f"{field} is invalid field. Please read document to check valid field."
            )

        j = daily_stock_t.join(
            security_t, daily_stock_t.c.I_SECURITY == security_t.c.I_SECURITY
        )

        stmt = (
            select(
                [
                    daily_stock_t.c.D_TRADE.label(TRADE_DATE),
                    func.trim(security_t.c.N_SECURITY).label(NAME),
                    value_col.label(VALUE),
                ]
            )
            .select_from(j)
            .order_by(func.DATE(daily_stock_t.c.D_TRADE))
        )
        if "I_TRADING_METHOD" in daily_stock_t.c:
            stmt = stmt.where(
                func.trim(daily_stock_t.c.I_TRADING_METHOD) == "A"
            )  # Auto Matching

        vld.check_start_end_date(
            start_date=start_date,
            end_date=end_date,
            last_update_date=self.last_table_update(daily_stock_t.name),
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=daily_stock_t.c.D_TRADE,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt, index_col=TRADE_DATE)

        df = self._pivot_name_value(df)

        if field in {
            fld.D_PRIOR,
            fld.D_OPEN,
            fld.D_HIGH,
            fld.D_LOW,
            fld.D_CLOSE,
            fld.D_AVERAGE,
            fld.D_LAST_BID,
            fld.D_LAST_OFFER,
            fld.D_DPS,
            fld.D_EPS,
            fld.D_ACC_DPS,
        }:
            df = self._merge_adjust_factor(
                df, is_multiply=True, adjusted_list=adjusted_list
            )
        elif field in {
            fld.D_VOLUME,
            fld.D_TOTAL_VOLUME,
        }:
            df = self._merge_adjust_factor(
                df, is_multiply=False, adjusted_list=adjusted_list
            )

        if symbol_list != None:
            df = df.reindex(columns=[i for i in symbol_list if i in df.columns])  # type: ignore

        return df

    def get_data_symbol_quarterly(
        self,
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from tables FINANCIAL_STAT_STD and FINANCIAL_SCREEN.If field is
        in both table, the data from FINANCIAL_STAT_STD will be used.

        FINANCIAL_STAT_STD using data from column M_ACCOUNT. FINANCIAL_SCREEN filter by I_PERIOD_TYPE='QY' and I_PERIOD in ('Q1','Q2','Q3','Q4').

        Index date is trade date (DAILY_STOCK_STAT.D_TRADE). Data is showing at first trade date which join on D_AS_OF.

        Parameters
        ----------
        field : str
            - year
            - period_type
            - period
            - quarter
            - accumulate
            - as_of
            - total_asset
            - total_liability
            - shld_equity
            - total_revenue
            - total_expense
            - net_profit
            - eps
            - de
            - net_profit_margin
            - gross_profit_margin
            - roa
            - roe
            - asset_turnover
            - ebit
            - fix_asset_turnover
            - current_ratio
            - quick_ratio
            - interest_coverage
            - ar_turnover
            - inventory_turnover
            - ap_turnover
            - cash_cycle
            - ebitda
            - net_operating
            - net_investing
            - net_financing
            - net_cashflow
            - dscr
            - ibde
            - account_payable
            - account_receivable
            - accrued_int_receive
            - allowance
            - cap_paidin
            - cap_paidup
            - cash
            - common_share
            - current_asset
            - current_liability
            - deposit
            - earning_asset
            - int_bearing_debt
            - inventory
            - invest_asset
            - investment
            - invest_security
            - loan
            - loan_from_relatedparty
            - loan_revenue
            - loan_to_relatedparty
            - longterm_liability_currentportion
            - longterm_liability_net_currentportion
            - minority_interest
            - ppe
            - preferred_share
            - retain_earning
            - retain_earning_unappropriate
            - shld_equity
            - short_invest
            - total_asset
            - total_equity
            - total_liability
            - change_ppe
            - dividend
            - dp
            - net_cash_flow
            - net_financing
            - net_investing
            - net_operating
            - bad_debt
            - broker_fee
            - cos
            - ebit
            - ebitda
            - ebt
            - int_dvd_income
            - interest_expense
            - interest_income
            - invest_sec_rev
            - loan_deposit_revenue
            - net_premium
            - net_profit
            - net_profit_incl_minority
            - net_profit_ordinary
            - operating_expense
            - operating_revenue
            - pl_other_activities
            - sale
            - selling_admin
            - selling_admin_exc_renumuration
            - total_expense
            - total_revenue
            - eps
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of trade date (DAILY_STOCK_STAT.D_TRADE), by default None
        end_date : Optional[str]
            end of trade date (DAILY_STOCK_STAT.D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - symbol(N_SECURITY): str as column
            - trade date(DAILY_STOCK_STAT.D_TRADE): date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_symbol_quarterly(
        ...     field=fld.Q_TOTAL_REVENUE,
        ...     symbol_list=["COM7", "MALEE"],
        ...     start_date="2022-02-01",
        ...     end_date=None,
        ... )
                           COM7      MALEE
        2022-03-01          NaN  953995.79
        2022-03-04  17573710.66        NaN
        """
        return self._get_fundamental_data(
            symbol_list=symbol_list,
            field=field,
            start_date=start_date,
            end_date=end_date,
            timeframe=fld.TIMEFRAME_QUARTERLY,
        )

    def get_data_symbol_yearly(
        self,
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table FINANCIAL_STAT_STD and FINANCIAL_SCREEN. If field is
        in both table, the data from FINANCIAL_STAT_STD will be used.

        FINANCIAL_STAT_STD filter by "I_QUARTER"='9' and using data from column M_ACCOUNT. FINANCIAL_SCREEN filter by I_PERIOD_TYPE='QY' and I_PERIOD='YE'.

        Index date is trade date (DAILY_STOCK_STAT.D_TRADE). Data is showing at first trade date which join on D_AS_OF.

        Parameters
        ----------
        field : str
            - year
            - period_type
            - period
            - quarter
            - accumulate
            - as_of
            - total_asset
            - total_liability
            - shld_equity
            - total_revenue
            - total_expense
            - net_profit
            - eps
            - de
            - net_profit_margin
            - gross_profit_margin
            - roa
            - roe
            - asset_turnover
            - ebit
            - fix_asset_turnover
            - current_ratio
            - quick_ratio
            - interest_coverage
            - ar_turnover
            - inventory_turnover
            - ap_turnover
            - cash_cycle
            - ebitda
            - net_operating
            - net_investing
            - net_financing
            - net_cashflow
            - dscr
            - ibde
            - account_payable
            - account_receivable
            - accrued_int_receive
            - allowance
            - cap_paidin
            - cap_paidup
            - cash
            - common_share
            - current_asset
            - current_liability
            - deposit
            - earning_asset
            - int_bearing_debt
            - inventory
            - invest_asset
            - investment
            - invest_security
            - loan
            - loan_from_relatedparty
            - loan_revenue
            - loan_to_relatedparty
            - longterm_liability_currentportion
            - longterm_liability_net_currentportion
            - minority_interest
            - ppe
            - preferred_share
            - retain_earning
            - retain_earning_unappropriate
            - shld_equity
            - short_invest
            - total_asset
            - total_equity
            - total_liability
            - change_ppe
            - dividend
            - dp
            - net_cash_flow
            - net_financing
            - net_investing
            - net_operating
            - bad_debt
            - broker_fee
            - cos
            - ebit
            - ebitda
            - ebt
            - int_dvd_income
            - interest_expense
            - interest_income
            - invest_sec_rev
            - loan_deposit_revenue
            - net_premium
            - net_profit
            - net_profit_incl_minority
            - net_profit_ordinary
            - operating_expense
            - operating_revenue
            - pl_other_activities
            - sale
            - selling_admin
            - selling_admin_exc_renumuration
            - total_expense
            - total_revenue
            - eps
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of trade date (DAILY_STOCK_STAT.D_TRADE), by default None
        end_date : Optional[str]
            end of trade date (DAILY_STOCK_STAT.D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - symbol(N_SECURITY): str as column
            - trade date(DAILY_STOCK_STAT.D_TRADE): date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_symbol_yearly(
        ...     field=fld.Y_TOTAL_REVENUE,
        ...     symbol_list=["COM7", "MALEE"],
        ...     start_date="2022-02-01",
        ...     end_date=None,
        ... )
                           COM7       MALEE
        2022-03-01          NaN  3488690.79
        2022-03-04  51154660.73         NaN
        """
        return self._get_fundamental_data(
            symbol_list=symbol_list,
            field=field,
            start_date=start_date,
            end_date=end_date,
            timeframe=fld.TIMEFRAME_YEARLY,
        )

    def get_data_symbol_ttm(
        self,
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Trailing 12 months (TTM) is a term used to describe the past 12
        consecutive months of a company's performance data.

        TTM can be calculate only Income Statement and Cashflow, but not Financial Ratio and Balance Sheet.

        Data from table FINANCIAL_SCREEN.

        FINANCIAL_STAT_STD filter by using data from column M_ACC_ACCOUNT_12M.

        Index date is trade date (DAILY_STOCK_STAT.D_TRADE). Data is showing at first trade date which join on D_AS_OF.

        Parameters
        ----------
        field : str
            - change_ppe
            - dividend
            - dp
            - net_cash_flow
            - net_financing
            - net_investing
            - net_operating
            - bad_debt
            - broker_fee
            - cos
            - ebit
            - ebitda
            - ebt
            - int_dvd_income
            - interest_expense
            - interest_income
            - invest_sec_rev
            - loan_deposit_revenue
            - net_premium
            - net_profit
            - net_profit_incl_minority
            - net_profit_ordinary
            - operating_expense
            - operating_revenue
            - pl_other_activities
            - sale
            - selling_admin
            - selling_admin_exc_renumuration
            - total_expense
            - total_revenue
            - eps
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of trade date (DAILY_STOCK_STAT.D_TRADE), by default None
        end_date : Optional[str]
            end of trade date (DAILY_STOCK_STAT.D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - symbol(N_SECURITY): str as column
            - trade date(DAILY_STOCK_STAT.D_TRADE): date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_symbol_ttm(
        ...     field=fld.Q_TOTAL_REVENUE,
        ...     symbol_list=["COM7", "MALEE"],
        ...     start_date="2022-02-01",
        ...     end_date=None,
        ... )
                           COM7       MALEE
        2022-03-01          NaN  3488690.79
        2022-03-04  51154660.73         NaN
        """
        return self._get_fundamental_data(
            symbol_list=symbol_list,
            field=field,
            start_date=start_date,
            end_date=end_date,
            timeframe=fld.TIMEFRAME_TTM,
        )

    def get_data_symbol_ytd(
        self,
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Year to date (YTD) refers to the period of time beginning the first
        day of the current calendar year or fiscal year up to the current date.

        Data from table FINANCIAL_STAT_STD and FINANCIAL_SCREEN. If field is in both table, the data from FINANCIAL_STAT_STD will be used.

        FINANCIAL_STAT_STD using data from column M_ACC_ACCOUNT. FINANCIAL_SCREEN filter by I_PERIOD_TYPE='QY' and I_PERIOD in ('Q1','6M','9M','YE').

        Index date is trade date (DAILY_STOCK_STAT.D_TRADE). Data is showing at first DAILY_STOCK_STAT.D_TRADE which join on D_AS_OF.

        Parameters
        ----------
        field : str
            - year
            - period_type
            - period
            - quarter
            - accumulate
            - as_of
            - total_asset
            - total_liability
            - shld_equity
            - total_revenue
            - total_expense
            - net_profit
            - eps
            - de
            - net_profit_margin
            - gross_profit_margin
            - roa
            - roe
            - asset_turnover
            - ebit
            - fix_asset_turnover
            - current_ratio
            - quick_ratio
            - interest_coverage
            - ar_turnover
            - inventory_turnover
            - ap_turnover
            - cash_cycle
            - ebitda
            - net_operating
            - net_investing
            - net_financing
            - net_cashflow
            - dscr
            - ibde
            - account_payable
            - account_receivable
            - accrued_int_receive
            - allowance
            - cap_paidin
            - cap_paidup
            - cash
            - common_share
            - current_asset
            - current_liability
            - deposit
            - earning_asset
            - int_bearing_debt
            - inventory
            - invest_asset
            - investment
            - invest_security
            - loan
            - loan_from_relatedparty
            - loan_revenue
            - loan_to_relatedparty
            - longterm_liability_currentportion
            - longterm_liability_net_currentportion
            - minority_interest
            - ppe
            - preferred_share
            - retain_earning
            - retain_earning_unappropriate
            - shld_equity
            - short_invest
            - total_asset
            - total_equity
            - total_liability
            - change_ppe
            - dividend
            - dp
            - net_cash_flow
            - net_financing
            - net_investing
            - net_operating
            - bad_debt
            - broker_fee
            - cos
            - ebit
            - ebitda
            - ebt
            - int_dvd_income
            - interest_expense
            - interest_income
            - invest_sec_rev
            - loan_deposit_revenue
            - net_premium
            - net_profit
            - net_profit_incl_minority
            - net_profit_ordinary
            - operating_expense
            - operating_revenue
            - pl_other_activities
            - sale
            - selling_admin
            - selling_admin_exc_renumuration
            - total_expense
            - total_revenue
            - eps
        symbol_list : Optional[List[str]]
            N_SECURITY in symbol_list, must be unique, by default None
        start_date : Optional[str]
            start of trade date (DAILY_STOCK_STAT.D_TRADE), by default None
        end_date : Optional[str]
            end of trade date (DAILY_STOCK_STAT.D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - symbol(N_SECURITY): str as column
            - trade date(DAILY_STOCK_STAT.D_TRADE): date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_symbol_ytd(
        ...     field=fld.Q_TOTAL_REVENUE,
        ...     symbol_list=["COM7", "MALEE"],
        ...     start_date="2022-02-01",
        ...     end_date=None,
        ... )
                           COM7       MALEE
        2022-03-01          NaN  3488690.79
        2022-03-04  51154660.73         NaN
        """
        return self._get_fundamental_data(
            symbol_list=symbol_list,
            field=field,
            start_date=start_date,
            end_date=end_date,
            timeframe=fld.TIMEFRAME_YTD,
        )

    def get_data_index_daily(
        self,
        field: str,
        index_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table MKTSTAT_DAILY_INDEX, MKTSTAT_DAILY_MARKET.

        Parameters
        ----------
        field : str
            - high
            - low
            - close
            - tri
            - total_trans
            - total_volume
            - total_value
            - mkt_pe
            - mkt_pbv
            - mkt_yield
            - mkt_cap
            - mkt_par_value
            - trading_day
            - new_company
            - delisted_company
            - move_in_company
            - move_out_company
            - listed_company
            - listed_stock
        index_list : Optional[List[str]]
            N_SECTOR in index_list, by default None. More index can be found in ezyquant.fields
        start_date : Optional[str]
            start of trade_date (D_TRADE), by default None
        end_date : Optional[str]
            end of trade_date (D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - index: str as column
            - trade_date: date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_index_daily(
        ...     field=fld.D_INDEX_CLOSE,
        ...     index_list=[fld.INDEX_SET, fld.INDEX_SET100],
        ...     start_date="2022-01-01",
        ...     end_date="2022-01-10",
        ... )
                        SET   SET100
        2022-01-04  1670.28  2283.56
        2022-01-05  1676.79  2291.71
        2022-01-06  1653.03  2251.78
        2022-01-07  1657.62  2257.40
        2022-01-10  1657.06  2256.14
        """
        field = field.lower()

        sector_t = self._table("SECTOR")

        if field in fld.MKTSTAT_DAILY_INDEX_MAP:
            mktstat_daily_t = self._table("MKTSTAT_DAILY_INDEX")
            field_col = mktstat_daily_t.c[fld.MKTSTAT_DAILY_INDEX_MAP[field]]
        elif field in fld.MKTSTAT_DAILY_MARKET_MAP:
            mktstat_daily_t = self._table("MKTSTAT_DAILY_MARKET")
            field_col = mktstat_daily_t.c[fld.MKTSTAT_DAILY_MARKET_MAP[field]]
        else:
            raise InputError(
                f"{field} is invalid field. Please read document to check valid field."
            )

        j = self._join_sector_table(mktstat_daily_t)

        stmt = (
            select(
                [
                    mktstat_daily_t.c.D_TRADE.label(TRADE_DATE),
                    func.trim(sector_t.c.N_SECTOR).label(NAME),
                    field_col.label(VALUE),
                ]
            )
            .select_from(j)
            .order_by(func.DATE(mktstat_daily_t.c.D_TRADE))
        )

        vld.check_start_end_date(
            start_date=start_date,
            end_date=end_date,
            last_update_date=self.last_table_update(mktstat_daily_t.name),
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=sector_t.c.N_SECTOR,
            date_column=mktstat_daily_t.c.D_TRADE,
            symbol_list=index_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt, index_col=TRADE_DATE)

        df = self._pivot_name_value(df)

        return df

    def get_data_sector_daily(
        self,
        field: str,
        market: str = fld.MARKET_SET,
        sector_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table DAILY_SECTOR_INFO. Filter only sector data.

        Parameters
        ----------
        field : str
            - prior
            - open
            - high
            - low
            - close
            - trans
            - volume
            - value
            - mkt_pe
            - mkt_pbv
            - mkt_yield
            - mkt_cap
            - turnover
            - share_listed_avg
            - beta
            - turnover_volume
            - 12m_dvd_yield
        market : str
            I_MARKET e.g. 'SET', 'mai', by default 'SET'
        sector_list : Optional[List[str]]
            N_SECTOR in sector_list, by default None. More sector can be found in ezyquant.fields
        start_date : Optional[str]
            start of trade_date (D_TRADE), by default None
        end_date : Optional[str]
            end of trade_date (D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - sector: str as column
            - trade_date: date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_sector_daily(
        ...     field=fld.D_SECTOR_CLOSE,
        ...     sector_list=[fld.SECTOR_AGRI, fld.SECTOR_BANK],
        ...     start_date="2022-01-01",
        ...     end_date="2022-01-10",
        ... )
                      AGRI    BANK
        2022-01-04  296.13  421.31
        2022-01-05  297.66  423.08
        2022-01-06  299.85  417.30
        2022-01-07  300.12  421.00
        2022-01-10  306.93  423.81
        """
        return self._get_daily_sector_info(
            field=field,
            sector_list=sector_list,
            start_date=start_date,
            end_date=end_date,
            market=market,
            f_data="S",  # sector
        )

    def get_data_industry_daily(
        self,
        field: str,
        market: str = fld.MARKET_SET,
        industry_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Data from table DAILY_SECTOR_INFO. Filter only industry data.

        Parameters
        ----------
        field : str
            - prior
            - open
            - high
            - low
            - close
            - trans
            - volume
            - value
            - mkt_pe
            - mkt_pbv
            - mkt_yield
            - mkt_cap
            - turnover
            - share_listed_avg
            - beta
            - turnover_volume
            - 12m_dvd_yield
        market : str
            I_MARKET e.g. 'SET', 'mai', by default 'SET'
        industry_list : Optional[List[str]]
            N_SECTOR in industry_list, by default None. More industry can be found in ezyquant.fields
        start_date : Optional[str]
            start of trade_date (D_TRADE), by default None
        end_date : Optional[str]
            end of trade_date (D_TRADE), by default None

        Returns
        -------
        pd.DataFrame
            - industry: str as column
            - trade_date: date as index

        Examples
        --------
        >>> from ezyquant import SETDataReader
        >>> from ezyquant import fields as fld
        >>> sdr = SETDataReader()
        >>> sdr.get_data_industry_daily(
        ...     field=fld.D_INDUSTRY_CLOSE,
        ...     industry_list=[fld.INDUSTRY_AGRO, fld.INDUSTRY_CONSUMP],
        ...     start_date="2022-01-01",
        ...     end_date="2022-01-10",
        ... )
                      AGRO  CONSUMP
        2022-01-04  485.98    92.55
        2022-01-05  484.98    93.21
        2022-01-06  482.90    92.96
        2022-01-07  484.50    93.04
        2022-01-10  487.10    93.97
        """
        return self._get_daily_sector_info(
            field=field,
            sector_list=industry_list,
            start_date=start_date,
            end_date=end_date,
            market=market,
            f_data="I",  # industry
        )

    """
    Protected methods
    """

    def _table(self, name: str) -> Table:
        return Table(name, self._metadata, autoload=True)

    def _execute(self, stmt: Select):
        return self._engine.execute(stmt)

    def _read_sql_query(
        self, stmt: Select, index_col: Optional[str] = None
    ) -> pd.DataFrame:
        col_name_list = [i.name for i in stmt.selected_columns]

        parse_dates = [i for i in col_name_list if i.endswith("_date")]

        df = pd.read_sql_query(
            stmt,
            self._engine,
            index_col=index_col,
            parse_dates=parse_dates,
        )

        if VALUE in col_name_list:
            try:
                df = df.astype({VALUE: "float64"})
            except ValueError:
                pass

        return df

    def _filter_stmt_by_date(
        self,
        stmt: Select,
        column: Column,
        start_date: Optional[str],
        end_date: Optional[str],
    ):
        vld.check_start_end_date(start_date, end_date)

        if start_date != None:
            stmt = stmt.where(func.DATE(column) >= func.DATE(start_date))
        if end_date != None:
            stmt = stmt.where(func.DATE(column) <= func.DATE(end_date))

        return stmt

    def _filter_str_in_list(
        self, stmt: Select, column: Column, values: Optional[List[str]]
    ):
        vld.check_duplicate(values)
        if values != None:
            values = [i.strip().upper() for i in values]
            stmt = stmt.where(func.upper(func.trim(column)).in_(values))
        return stmt

    def _filter_stmt_by_symbol_and_date(
        self,
        stmt: Select,
        symbol_column: Column,
        date_column: Column,
        symbol_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
    ):
        stmt = self._filter_str_in_list(
            stmt=stmt, column=symbol_column, values=symbol_list
        )
        stmt = self._filter_stmt_by_date(
            stmt=stmt, column=date_column, start_date=start_date, end_date=end_date
        )
        return stmt

    @lru_cache
    def _d_trade_subquery(self):
        daily_stock_stat_t = self._table("DAILY_STOCK_STAT")
        return (
            select(
                [
                    daily_stock_stat_t.c.I_SECURITY,
                    daily_stock_stat_t.c.D_AS_OF,
                    func.min(func.DATE(daily_stock_stat_t.c.D_TRADE)).label("D_TRADE"),
                ]
            )
            .group_by(
                daily_stock_stat_t.c.I_SECURITY, func.DATE(daily_stock_stat_t.c.D_AS_OF)
            )
            .subquery()
        )

    def _join_security_and_d_trade_subquery(self, table: Table):
        security_t = self._table("SECURITY")
        d_trade_subquery = self._d_trade_subquery()
        return table.join(
            security_t, table.c.I_SECURITY == security_t.c.I_SECURITY
        ).join(
            d_trade_subquery,
            and_(
                table.c.I_SECURITY == d_trade_subquery.c.I_SECURITY,
                func.DATE(table.c.D_AS_OF) == func.DATE(d_trade_subquery.c.D_AS_OF),
            ),
        )

    def _join_sector_table(self, table: Table, isouter: bool = False):
        sector_t = self._table("SECTOR")
        return table.join(
            sector_t,
            and_(
                table.c.I_MARKET == sector_t.c.I_MARKET,
                table.c.I_INDUSTRY == sector_t.c.I_INDUSTRY,
                table.c.I_SECTOR == sector_t.c.I_SECTOR,
                table.c.I_SUBSECTOR == sector_t.c.I_SUBSECTOR,
            ),
            isouter=isouter,
        )

    def _merge_adjust_factor(
        self,
        df: pd.DataFrame,
        is_multiply: bool = True,
        start_adjust_date: Optional[str] = None,
        adjusted_list: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Adjust dataframe by adjust_factor.

        Parameters
        ----------
        df : pd.DataFrame
            index must be trade_date and columns must be symbol
        is_multiply : bool, optional
            is multiply by adjust factor (adjust factor usually less than 1), by default True
        start_adjust_date : Optional[str], optional
            start of get adjust factor effect date, by default None
        adjusted_list : Optional[List[str]], optional
            n_ca_type list, by default None

        Returns
        -------
        pd.DataFrame
            index is trade_date and columns is symbol
        """
        if df.empty:
            return df

        if start_adjust_date == None:
            start_adjust_date = utils.date_to_str(df.index.min())

        symbol_list = df.columns.tolist()

        adjust_factor_df = self._get_pivot_adjust_factor_df(
            min_date=utils.date_to_str(df.index.min()),
            max_date=utils.date_to_str(df.index.max()),
            symbol_list=symbol_list,
            start_date=start_adjust_date,
            ca_type_list=adjusted_list,
        )

        # reindex adjust_factor_df to df
        adjust_factor_df = adjust_factor_df.reindex(df.index)  # type: ignore

        # multiply or divide
        if is_multiply:
            df = df * adjust_factor_df
        else:
            df = df / adjust_factor_df

        return df

    def _merge_adjust_factor_dividend(
        self,
        df: pd.DataFrame,
        start_adjust_date: Optional[str] = None,
        adjusted_list: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        if df.empty:
            return df

        if start_adjust_date == None:
            start_adjust_date = utils.date_to_str(df["ex_date"].min())

        symbol_list = df["symbol"].unique().tolist()

        adjust_factor_df = self._get_pivot_adjust_factor_df(
            min_date=utils.date_to_str(df["ex_date"].min()),
            max_date=utils.date_to_str(df["ex_date"].max()),
            symbol_list=symbol_list,
            start_date=start_adjust_date,
            ca_type_list=adjusted_list,
        )

        adjust_factor_df = (
            adjust_factor_df.stack().rename("adjust_factor").reset_index()  # type: ignore
        )

        df = df.merge(
            adjust_factor_df,
            how="left",
            left_on=["ex_date", "symbol"],
            right_on=["effect_date", "symbol"],
            validate="m:1",
        )
        df["adjust_factor"] = df["adjust_factor"].fillna(1)

        df["dps"] *= df["adjust_factor"]

        df = df.drop(columns=["effect_date", "adjust_factor"])

        return df

    def _get_pivot_adjust_factor_df(
        self,
        min_date: str,
        max_date: str,
        symbol_list: List[str],
        start_date: Optional[str] = None,
        ca_type_list: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        df = self.get_adjust_factor(
            symbol_list=symbol_list if len(symbol_list) < 100 else None,
            start_date=start_date,
            ca_type_list=ca_type_list,
        )

        # pivot table
        df = df.pivot(index="effect_date", columns="symbol", values="adjust_factor")
        # reverse cumulate product adjust factor
        df = df.iloc[::-1].cumprod(skipna=True).iloc[::-1]

        # reindex trade date
        if not df.empty:
            max_date = max(max_date, utils.date_to_str(df.index.max()))
        df = df.reindex(
            index=pd.date_range(
                start=min_date,
                end=max_date,
                normalize=True,
                name="effect_date",
            ),  # type: ignore
            columns=symbol_list,  # type: ignore
        )

        # shift back 1 day
        df = df.shift(-1)

        # back fill and fill 1
        df = df.fillna(method="backfill").fillna(1)

        return df

    def _get_financial_screen_stmt(self, timeframe: str, field: str) -> Select:
        TIMEFRAME_MAP = {
            fld.TIMEFRAME_QUARTERLY: ("Q1", "Q2", "Q3", "Q4"),
            fld.TIMEFRAME_YEARLY: ("YE",),
            fld.TIMEFRAME_YTD: ("Q1", "6M", "9M", "YE"),
        }

        financial_screen_t = self._table("FINANCIAL_SCREEN")
        security_t = self._table("SECURITY")
        d_trade_subquery = self._d_trade_subquery()

        value_column = financial_screen_t.c[fld.FINANCIAL_SCREEN_MAP[field]]

        stmt = (
            select(
                [
                    d_trade_subquery.c.D_TRADE.label(TRADE_DATE),
                    func.trim(security_t.c.N_SECURITY).label(NAME),
                    value_column.label(VALUE),
                ]
            )
            .select_from(self._join_security_and_d_trade_subquery(financial_screen_t))
            .where(func.trim(financial_screen_t.c.I_PERIOD_TYPE) == "QY")
            .where(
                func.trim(financial_screen_t.c.I_PERIOD).in_(TIMEFRAME_MAP[timeframe])
            )
        )

        return stmt

    def _get_financial_stat_std_stmt(self, timeframe: str, field: str) -> Select:
        financial_stat_std_t = self._table("FINANCIAL_STAT_STD")
        security_t = self._table("SECURITY")
        d_trade_subquery = self._d_trade_subquery()

        if timeframe == fld.TIMEFRAME_QUARTERLY:
            value_column = financial_stat_std_t.c["M_ACCOUNT"]
        elif timeframe == fld.TIMEFRAME_YEARLY:
            if field in fld.FINANCIAL_STAT_STD_MAP["B"]:
                value_column = financial_stat_std_t.c["M_ACCOUNT"]
            else:
                value_column = financial_stat_std_t.c["M_ACC_ACCOUNT_12M"]
        elif timeframe == fld.TIMEFRAME_YTD:
            value_column = financial_stat_std_t.c["M_ACC_ACCOUNT"]
        elif timeframe == "average":
            value_column = financial_stat_std_t.c["M_ACCOUNT_AVG"]
        elif timeframe == fld.TIMEFRAME_TTM:
            value_column = financial_stat_std_t.c["M_ACC_ACCOUNT_12M"]
        else:
            raise ValueError(f"{timeframe} is not a valid timeframe")

        stmt = (
            select(
                [
                    d_trade_subquery.c.D_TRADE.label(TRADE_DATE),
                    func.trim(security_t.c.N_SECURITY).label(NAME),
                    value_column.label(VALUE),
                ]
            )
            .select_from(self._join_security_and_d_trade_subquery(financial_stat_std_t))
            .where(
                func.trim(financial_stat_std_t.c.N_ACCOUNT)
                == fld.FINANCIAL_STAT_STD_MAP_COMPACT[field]
            )
        )

        if timeframe == fld.TIMEFRAME_YEARLY:
            stmt = stmt.where(financial_stat_std_t.c.I_QUARTER == 9)

        return stmt

    def _get_fundamental_data(
        self,
        timeframe: str,
        field: str,
        symbol_list: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        fillna_value=None,
    ) -> pd.DataFrame:
        timeframe = timeframe.lower()
        field = field.lower()

        security_t = self._table("SECURITY")
        d_trade_subquery = self._d_trade_subquery()

        if field in fld.FINANCIAL_STAT_STD_MAP_COMPACT:
            stmt = self._get_financial_stat_std_stmt(field=field, timeframe=timeframe)
        elif field in fld.FINANCIAL_SCREEN_MAP and timeframe in (
            fld.TIMEFRAME_QUARTERLY,
            fld.TIMEFRAME_YEARLY,
            fld.TIMEFRAME_YTD,
        ):
            stmt = self._get_financial_screen_stmt(field=field, timeframe=timeframe)
        else:
            raise InputError(
                f"{field} is invalid field. Please read document to check valid field."
            )

        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=d_trade_subquery.c.D_TRADE,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt)

        # duplicate key mostly I_ACCT_FORM 6,7
        df = df.drop_duplicates(subset=[TRADE_DATE, NAME], keep="last")
        df = df.set_index(TRADE_DATE)

        if fillna_value != None:
            df = df.fillna(fillna_value)

        df = self._pivot_name_value(df)

        return df

    def _get_daily_sector_info(
        self,
        field: str,
        sector_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
        market: str,
        f_data: str,
    ) -> pd.DataFrame:
        market = market.upper()
        field = field.lower()
        f_data = f_data.upper()

        assert f_data in {"I", "S"}, f"{f_data} is not a valid f_data"

        sector_t = self._table("SECTOR")
        daily_sector_info_t = self._table("DAILY_SECTOR_INFO")

        j = self._join_sector_table(daily_sector_info_t)

        try:
            field_col = daily_sector_info_t.c[fld.DAILY_SECTOR_INFO_MAP[field]]
        except KeyError:
            raise InputError(
                f"{field} is invalid field. Please read document to check valid field."
            )

        # N_SECTOR is the industry name in F_DATA = 'I'
        stmt = (
            select(
                [
                    daily_sector_info_t.c.D_TRADE.label(TRADE_DATE),
                    func.trim(sector_t.c.N_SECTOR).label(NAME),
                    field_col.label(VALUE),
                ]
            )
            .select_from(j)
            .where(sector_t.c.F_DATA == f_data)
            .where(sector_t.c.I_MARKET == fld.MARKET_MAP_UPPER[market])
            .where(sector_t.c.D_CANCEL == None)
            .order_by(func.DATE(daily_sector_info_t.c.D_TRADE))
        )

        vld.check_start_end_date(
            start_date=start_date,
            end_date=end_date,
            last_update_date=self.last_table_update(daily_sector_info_t.name),
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=sector_t.c.N_SECTOR,
            date_column=daily_sector_info_t.c.D_TRADE,
            symbol_list=sector_list,
            start_date=start_date,
            end_date=end_date,
        )

        df = self._read_sql_query(stmt, index_col=TRADE_DATE)

        df = self._pivot_name_value(df)

        return df

    def _get_daily_sector_info_by_security(
        self,
        field: str,
        symbol_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
        f_data: str,
    ) -> pd.DataFrame:
        field = field.lower()
        f_data = f_data.upper()

        assert f_data in {"I", "S"}, f"{f_data} is not a valid f_data"

        security_t = self._table("SECURITY")
        sector_t = self._table("SECTOR")
        daily_sector_info_t = self._table("DAILY_SECTOR_INFO")

        j = self._join_sector_table(daily_sector_info_t).join(
            security_t,
            and_(
                sector_t.c.I_MARKET == security_t.c.I_MARKET,
                sector_t.c.I_INDUSTRY == security_t.c.I_INDUSTRY,
                sector_t.c.I_SECTOR == security_t.c.I_SECTOR if f_data == "S" else True,
                sector_t.c.I_SUBSECTOR == security_t.c.I_SUBSECTOR,
            ),
        )

        try:
            field_col = daily_sector_info_t.c[fld.DAILY_SECTOR_INFO_MAP[field]]
        except KeyError:
            raise InputError(
                f"{field} is invalid field. Please read document to check valid field."
            )

        # N_SECTOR is the industry name in F_DATA = 'I'
        stmt = (
            select(
                [
                    daily_sector_info_t.c.D_TRADE.label(TRADE_DATE),
                    func.trim(security_t.c.N_SECURITY).label(NAME),
                    field_col.label(VALUE),
                ]
            )
            .select_from(j)
            .where(sector_t.c.F_DATA == f_data)
            .where(sector_t.c.D_CANCEL == None)
            .order_by(func.DATE(daily_sector_info_t.c.D_TRADE))
        )

        vld.check_start_end_date(
            start_date=start_date,
            end_date=end_date,
            last_update_date=self.last_table_update(daily_sector_info_t.name),
        )
        stmt = self._filter_stmt_by_symbol_and_date(
            stmt=stmt,
            symbol_column=security_t.c.N_SECURITY,
            date_column=daily_sector_info_t.c.D_TRADE,
            symbol_list=symbol_list,
            start_date=start_date,
            end_date=end_date,
        )
        df = self._read_sql_query(stmt, index_col=TRADE_DATE)

        df = self._pivot_name_value(df)

        return df

    def _get_last_as_of_date_in_security_index(
        self, current_date: Optional[str] = None
    ) -> Dict[str, str]:
        sector_t = self._table("SECTOR")
        security_index_t = self._table("SECURITY_INDEX")

        j = self._join_sector_table(security_index_t)
        stmt = (
            select(
                [
                    func.trim(sector_t.c.N_SECTOR),
                    func.max(func.DATE(security_index_t.c.D_AS_OF)),
                ]
            )
            .select_from(j)
            .group_by(sector_t.c.N_SECTOR)
        )
        stmt = self._filter_stmt_by_date(
            stmt=stmt,
            column=security_index_t.c.D_AS_OF,
            start_date=None,
            end_date=current_date,
        )

        result = self._execute(stmt).all()
        return dict(result)  # type: ignore

    """
    Custom business day functions
    """

    @lru_cache
    def _get_holidays(self) -> List[pd.Timestamp]:
        tds = self.get_trading_dates()
        return (
            pd.concat(
                [
                    pd.bdate_range(tds[0], tds[-1]).to_series(),
                    pd.DatetimeIndex(tds).to_series(),
                ]
            )
            .drop_duplicates(keep=False)
            .tolist()
        )

    def _SETBusinessDay(self, n: int = 1) -> CustomBusinessDay:
        return CustomBusinessDay(n, holidays=self._get_holidays())  # type: ignore

    """
    Static methods
    """

    @staticmethod
    def _pivot_name_value(df: pd.DataFrame) -> pd.DataFrame:
        df = utils.pivot_remove_index_name(df=df, columns=NAME, values=VALUE)
        return df


@lru_cache
def _SETDataReaderCached() -> SETDataReader:
    out: SETDataReader = utils.wrap_cache_class(SETDataReader)()  # type: ignore

    out.get_data_symbol_daily = utils.cache_dataframe_wrapper(out.get_data_symbol_daily)  # type: ignore
    out.get_data_symbol_quarterly = utils.cache_dataframe_wrapper(
        out.get_data_symbol_quarterly
    )
    out.get_data_symbol_yearly = utils.cache_dataframe_wrapper(
        out.get_data_symbol_yearly
    )
    out.get_data_symbol_ttm = utils.cache_dataframe_wrapper(out.get_data_symbol_ttm)
    out.get_data_symbol_ytd = utils.cache_dataframe_wrapper(out.get_data_symbol_ytd)

    return out


def SETBusinessDay(n: int = 1) -> CustomBusinessDay:
    return _SETDataReaderCached()._SETBusinessDay(n)



from dataclasses import dataclass
from datetime import datetime
from functools import cached_property


@dataclass(frozen=True)
class SETTrade:
    """SETTrade

    Parameters
    ----------
    matched_at : datetime
        The time when the trade was executed.
    symbol : str
        The symbol of the trade
    volume : float
        The volume of the trade, must be multiple of 100, positive is buy and negative is sell
    price : float
        The price of the trade
    pct_commission : float
        The percentage of commission, must be between 0 and 1
    """

    matched_at: datetime
    symbol: str
    volume: float
    price: float
    pct_commission: float

    def __post_init__(self):
        # matched_at
        assert isinstance(
            self.matched_at, datetime
        ), f"matched_at must be datetime, got {type(self.matched_at)}"

        # symbol
        assert isinstance(
            self.symbol, str
        ), f"symbol must be str, got {type(self.symbol)}"
        assert self.symbol != "", f"symbol must not be empty, got {self.symbol}"

        # volume
        assert self.volume != 0, f"volume must not be 0, got {self.volume}"
        assert (
            self.volume % 100 == 0
        ), f"volume must be multiple of 100, got {self.volume}"

        # price
        assert self.price > 0, f"price must be positive, got {self.price}"

        # pct_commission
        assert (
            0 <= self.pct_commission <= 1
        ), f"pct_commission must be between 0 and 1, got {self.pct_commission}"

    @cached_property
    def value(self) -> float:
        """Positive is Buy, Negative is Sell"""
        return self.price * self.volume

    @cached_property
    def commission(self) -> float:
        """Always positive"""
        return abs(self.value * self.pct_commission)

    @cached_property
    def value_with_commission(self) -> float:
        """Amount of cash reduced by this trade. Positive is Buy, Negative is Sell"""
        return self.value + self.commission
